from qhub.provider.cloud.commons import filter_by_highest_supported_k8s_version


def test_filter_by_highest_supported_k8s_version():
    k8s_versions = [
        "1.21.7",
        "1.21.9",
        "1.22.4",
        "1.22.6",
        "1.23.3",
        "1.23.5",
        "1.24.0",
    ]
    actual = filter_by_highest_supported_k8s_version(k8s_versions)
    expected = sorted(list(set(k8s_versions) - {"1.24.0"}))
    assert actual == expected



import os

# Disable history manager, we don't really use it and by default it
# puts an sqlite file on NFS, which is not something we want to do
c.Historymanager.enabled = False


# Change default umask for all subprocesses of the notebook server if
# set in the environment
if "NB_UMASK" in os.environ:
    os.umask(int(os.environ["NB_UMASK"], 8))



import functools
import json
import os

from aiohttp import web
from dask_gateway_server.auth import JupyterHubAuthenticator
from dask_gateway_server.options import Mapping, Options, Select


def dask_gateway_config(path="/var/lib/dask-gateway/config.json"):
    with open(path) as f:
        return json.load(f)


config = dask_gateway_config()


c.DaskGateway.log_level = config["gateway"]["loglevel"]

# Configure addresses
c.DaskGateway.address = ":8000"
c.KubeBackend.api_url = f'http://{config["gateway_service_name"]}.{config["gateway_service_namespace"]}:8000/api'

c.DaskGateway.backend_class = "dask_gateway_server.backends.kubernetes.KubeBackend"
c.KubeBackend.gateway_instance = config["gateway_service_name"]

# ========= Dask Cluster Default Configuration =========
c.KubeClusterConfig.image = (
    f"{config['cluster-image']['name']}:{config['cluster-image']['tag']}"
)
c.KubeClusterConfig.image_pull_policy = config["cluster"]["image_pull_policy"]
c.KubeClusterConfig.environment = config["cluster"]["environment"]
c.KubeClusterConfig.idle_timeout = config["cluster"]["idle_timeout"]

c.KubeClusterConfig.scheduler_cores = config["cluster"]["scheduler_cores"]
c.KubeClusterConfig.scheduler_cores_limit = config["cluster"]["scheduler_cores_limit"]
c.KubeClusterConfig.scheduler_memory = config["cluster"]["scheduler_memory"]
c.KubeClusterConfig.scheduler_memory_limit = config["cluster"]["scheduler_memory_limit"]
c.KubeClusterConfig.scheduler_extra_container_config = config["cluster"][
    "scheduler_extra_container_config"
]
c.KubeClusterConfig.scheduler_extra_pod_config = config["cluster"][
    "scheduler_extra_pod_config"
]

c.KubeClusterConfig.worker_cores = config["cluster"]["worker_cores"]
c.KubeClusterConfig.worker_cores_limit = config["cluster"]["worker_cores_limit"]
c.KubeClusterConfig.worker_memory = config["cluster"]["worker_memory"]
c.KubeClusterConfig.worker_memory_limit = config["cluster"]["worker_memory_limit"]
c.KubeClusterConfig.worker_extra_container_config = config["cluster"][
    "worker_extra_container_config"
]
c.KubeClusterConfig.worker_extra_pod_config = config["cluster"][
    "worker_extra_pod_config"
]


# ============ Authentication =================
class QHubAuthentication(JupyterHubAuthenticator):
    async def authenticate(self, request):
        user = await super().authenticate(request)
        url = f"{self.jupyterhub_api_url}/users/{user.name}"
        kwargs = {
            "headers": {"Authorization": "token %s" % self.jupyterhub_api_token},
            "ssl": self.ssl_context,
        }
        resp = await self.session.get(url, **kwargs)
        data = (await resp.json())["auth_state"]["oauth_user"]

        if (
            "dask_gateway_developer" not in data["roles"]
            and "dask_gateway_admin" not in data["roles"]
        ):
            raise web.HTTPInternalServerError(
                reason="Permission failure user does not have required dask_gateway roles"
            )

        user.admin = "dask_gateway_admin" in data["roles"]
        user.groups = [os.path.basename(group) for group in data["groups"]]
        return user


c.DaskGateway.authenticator_class = QHubAuthentication
c.JupyterHubAuthenticator.jupyterhub_api_url = config["jupyterhub_api_url"]
c.JupyterHubAuthenticator.jupyterhub_api_token = config["jupyterhub_api_token"]


# ==================== Profiles =======================
def get_packages(conda_prefix):
    try:
        packages = set()
        for filename in os.listdir(os.path.join(conda_prefix, "conda-meta")):
            if filename.endswith(".json"):
                with open(os.path.join(conda_prefix, "conda-meta", filename)) as f:
                    packages.add(json.load(f).get("name"))
        return packages
    except OSError as e:
        import logging

        logger = logging.getLogger()
        logger.error(f"An issue with a conda environment was encountered.\n{e}")


def get_conda_prefixes(conda_store_mount):
    for namespace in os.listdir(conda_store_mount):
        if os.path.isdir(os.path.join(conda_store_mount, namespace, "envs")):
            for name in os.listdir(os.path.join(conda_store_mount, namespace, "envs")):
                yield namespace, name, os.path.join(
                    conda_store_mount, namespace, "envs", name
                )


def list_dask_environments(conda_store_mount):
    for namespace, name, conda_prefix in get_conda_prefixes(conda_store_mount):
        packages = get_packages(conda_prefix)
        if packages and {"dask", "distributed"} <= packages:
            yield namespace, name, conda_prefix


def base_node_group():
    worker_node_group = {
        config["worker-node-group"]["key"]: config["worker-node-group"]["value"]
    }

    return {
        "scheduler_extra_pod_config": {"nodeSelector": worker_node_group},
        "worker_extra_pod_config": {"nodeSelector": worker_node_group},
    }


def base_conda_store_mounts(namespace, name):
    conda_store_pvc_name = config["conda-store-pvc"]
    conda_store_mount = config["conda-store-mount"]

    return {
        "scheduler_extra_pod_config": {
            "volumes": [
                {
                    "name": "conda-store",
                    "persistentVolumeClaim": {
                        "claimName": conda_store_pvc_name,
                    },
                }
            ]
        },
        "scheduler_extra_container_config": {
            "volumeMounts": [
                {
                    "mountPath": os.path.join(conda_store_mount, namespace),
                    "name": "conda-store",
                    "subPath": namespace,
                }
            ]
        },
        "worker_extra_pod_config": {
            "volumes": [
                {
                    "name": "conda-store",
                    "persistentVolumeClaim": {
                        "claimName": conda_store_pvc_name,
                    },
                }
            ]
        },
        "worker_extra_container_config": {
            "volumeMounts": [
                {
                    "mountPath": os.path.join(conda_store_mount, namespace),
                    "name": "conda-store",
                    "subPath": namespace,
                }
            ]
        },
        "worker_cmd": "/opt/conda-run-worker",
        "scheduler_cmd": "/opt/conda-run-scheduler",
        "environment": {
            "CONDA_ENVIRONMENT": os.path.join(
                conda_store_mount, namespace, "envs", name
            ),
        },
    }


def base_username_mount(username, uid=1000, gid=100):
    return {
        "scheduler_extra_pod_config": {"volumes": [{"name": "home", "emptyDir": {}}]},
        "scheduler_extra_container_config": {
            "securityContext": {"runAsUser": uid, "runAsGroup": gid, "fsGroup": gid},
            "workingDir": f"/home/{username}",
            "volumeMounts": [
                {
                    "mountPath": f"/home/{username}",
                    "name": "home",
                }
            ],
        },
        "worker_extra_pod_config": {"volumes": [{"name": "home", "emptyDir": {}}]},
        "worker_extra_container_config": {
            "securityContext": {"runAsUser": uid, "runAsGroup": gid, "fsGroup": gid},
            "workingDir": f"/home/{username}",
            "volumeMounts": [
                {
                    "mountPath": f"/home/{username}",
                    "name": "home",
                }
            ],
        },
        "environment": {
            "HOME": f"/home/{username}",
        },
    }


def worker_profile(options, user):
    namespace, name = options.conda_environment.split("/")
    return functools.reduce(
        deep_merge,
        [
            base_node_group(),
            base_conda_store_mounts(namespace, name),
            base_username_mount(user.name),
            config["profiles"][options.profile],
            {"environment": {**options.environment_vars}},
        ],
        {},
    )


def user_options(user):
    allowed_namespaces = set(["filesystem", "default", user.name] + list(user.groups))
    environments = {
        f"{namespace}/{name}": conda_prefix
        for namespace, name, conda_prefix in list_dask_environments(
            config["conda-store-mount"]
        )
        if namespace in allowed_namespaces
    }

    args = []
    if environments:
        args += [
            Select(
                "conda_environment",
                list(environments.keys()),
                default=list(environments.keys())[0],
                label="Environment",
            )
        ]
    if config["profiles"]:
        args += [
            Select(
                "profile",
                list(config["profiles"].keys()),
                default=list(config["profiles"].keys())[0],
                label="Cluster Profile",
            )
        ]

    args += [
        Mapping("environment_vars", {}, label="Environment Variables"),
    ]

    return Options(
        *args,
        handler=worker_profile,
    )


c.Backend.cluster_options = user_options


# ============== utils ============
def deep_merge(d1, d2):
    """Deep merge two dictionaries.
    >>> value_1 = {
    'a': [1, 2],
    'b': {'c': 1, 'z': [5, 6]},
    'e': {'f': {'g': {}}},
    'm': 1,
    }

    >>> value_2 = {
        'a': [3, 4],
        'b': {'d': 2, 'z': [7]},
        'e': {'f': {'h': 1}},
        'm': [1],
    }

    >>> print(deep_merge(value_1, value_2))
    {'m': 1, 'e': {'f': {'g': {}, 'h': 1}}, 'b': {'d': 2, 'c': 1, 'z': [5, 6, 7]}, 'a': [1, 2, 3,  4]}
    """
    if isinstance(d1, dict) and isinstance(d2, dict):
        d3 = {}
        for key in d1.keys() | d2.keys():
            if key in d1 and key in d2:
                d3[key] = deep_merge(d1[key], d2[key])
            elif key in d1:
                d3[key] = d1[key]
            elif key in d2:
                d3[key] = d2[key]
        return d3
    elif isinstance(d1, list) and isinstance(d2, list):
        return [*d1, *d2]
    else:  # if they don't match use left one
        return d1



import contextlib
import functools
import os
import pathlib
import re
import signal
import subprocess
import sys
import threading
import time
from typing import Dict, List

import requests
from requests.exceptions import ConnectionError
from ruamel.yaml import YAML

from qhub.constants import DEFAULT_QHUB_DASK_VERSION, DEFAULT_QHUB_IMAGE_TAG
from qhub.provider.cicd.github import get_latest_repo_tag
from qhub.provider.cloud import (
    amazon_web_services,
    azure_cloud,
    digital_ocean,
    google_cloud,
)

# environment variable overrides
QHUB_K8S_VERSION = os.getenv("QHUB_K8S_VERSION", None)
QHUB_GH_BRANCH = os.getenv("QHUB_GH_BRANCH", None)
QHUB_IMAGE_TAG = os.getenv("QHUB_IMAGE_TAG", None)
QHUB_DASK_VERSION = os.getenv("QHUB_DASK_VERSION", None)

DO_ENV_DOCS = (
    "https://docs.qhub.dev/en/stable/source/installation/setup.html#digital-ocean"
)
AWS_ENV_DOCS = "https://docs.qhub.dev/en/stable/source/installation/setup.html#amazon-web-services-aws"
GCP_ENV_DOCS = "https://docs.qhub.dev/en/stable/source/installation/setup.html#google-cloud-platform"
AZURE_ENV_DOCS = (
    "https://docs.qhub.dev/en/stable/source/installation/setup.html#microsoft-azure"
)

CONDA_FORGE_CHANNEL_DATA_URL = "https://conda.anaconda.org/conda-forge/channeldata.json"

DOCKER_IMAGE_OWNER = "nebari-dev"
DOCKER_IMAGE_REPO = "nebari-docker-images"

# Regex for suitable project names
namestr_regex = r"^[A-Za-z][A-Za-z\-_]*[A-Za-z]$"

# Create a ruamel object with our favored config, for universal use
yaml = YAML()
yaml.preserve_quotes = True
yaml.default_flow_style = False


@contextlib.contextmanager
def timer(logger, prefix):
    start_time = time.time()
    yield
    logger.info(f"{prefix} took {time.time() - start_time:.3f} [s]")


@contextlib.contextmanager
def change_directory(directory):
    current_directory = os.getcwd()
    os.chdir(directory)
    yield
    os.chdir(current_directory)


def run_subprocess_cmd(processargs, **kwargs):
    """Runs subprocess command with realtime stdout logging with optional line prefix."""
    if "prefix" in kwargs:
        line_prefix = f"[{kwargs['prefix']}]: ".encode("utf-8")
        kwargs.pop("prefix")
    else:
        line_prefix = b""

    timeout = 0
    if "timeout" in kwargs:
        timeout = kwargs.pop("timeout")  # in seconds

    strip_errors = kwargs.pop("strip_errors", False)

    process = subprocess.Popen(
        processargs,
        **kwargs,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        preexec_fn=os.setsid,
    )
    # Set timeout thread
    timeout_timer = None
    if timeout > 0:

        def kill_process():
            try:
                os.killpg(process.pid, signal.SIGTERM)
            except ProcessLookupError:
                pass  # Already finished

        timeout_timer = threading.Timer(timeout, kill_process)
        timeout_timer.start()

    for line in iter(lambda: process.stdout.readline(), b""):
        full_line = line_prefix + line
        if strip_errors:
            full_line = full_line.decode("utf-8")
            full_line = re.sub(
                r"\x1b\[31m", "", full_line
            )  # Remove red ANSI escape code
            full_line = full_line.encode("utf-8")

        sys.stdout.buffer.write(full_line)
        sys.stdout.flush()

    if timeout_timer is not None:
        timeout_timer.cancel()

    return process.wait(
        timeout=10
    )  # Should already have finished because we have drained stdout


def check_cloud_credentials(config):
    if config["provider"] == "gcp":
        for variable in {"GOOGLE_CREDENTIALS"}:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {GCP_ENV_DOCS}"""
                )
    elif config["provider"] == "azure":
        for variable in {
            "ARM_CLIENT_ID",
            "ARM_CLIENT_SECRET",
            "ARM_SUBSCRIPTION_ID",
            "ARM_TENANT_ID",
        }:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {AZURE_ENV_DOCS}"""
                )
    elif config["provider"] == "aws":
        for variable in {
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
        }:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {AWS_ENV_DOCS}"""
                )
    elif config["provider"] == "do":
        for variable in {
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "SPACES_ACCESS_KEY_ID",
            "SPACES_SECRET_ACCESS_KEY",
            "DIGITALOCEAN_TOKEN",
        }:
            if variable not in os.environ:
                raise ValueError(
                    f"""Missing the following required environment variable: {variable}\n
                    Please see the documentation for more information: {DO_ENV_DOCS}"""
                )

        if os.environ["AWS_ACCESS_KEY_ID"] != os.environ["SPACES_ACCESS_KEY_ID"]:
            raise ValueError(
                f"""The environment variables AWS_ACCESS_KEY_ID and SPACES_ACCESS_KEY_ID must be equal\n
                See {DO_ENV_DOCS} for more information"""
            )

        if (
            os.environ["AWS_SECRET_ACCESS_KEY"]
            != os.environ["SPACES_SECRET_ACCESS_KEY"]
        ):
            raise ValueError(
                f"""The environment variables AWS_SECRET_ACCESS_KEY and SPACES_SECRET_ACCESS_KEY must be equal\n
                See {DO_ENV_DOCS} for more information"""
            )
    elif config["provider"] in ["local", "existing"]:
        pass
    else:
        raise ValueError("Cloud Provider configuration not supported")


def load_yaml(config_filename: pathlib.Path):
    """
    Return yaml dict containing config loaded from config_filename.
    """
    with config_filename.open() as f:
        config = yaml.load(f.read())

    return config


def backup_config_file(filename: pathlib.Path, extrasuffix: str = ""):
    if not filename.exists():
        return

    # Backup old file
    backup_filename = pathlib.Path(f"{filename}{extrasuffix}.backup")

    if backup_filename.exists():
        i = 1
        while True:
            next_backup_filename = pathlib.Path(f"{backup_filename}~{i}")
            if not next_backup_filename.exists():
                backup_filename = next_backup_filename
                break
            i = i + 1

    filename.rename(backup_filename)
    print(f"Backing up {filename} as {backup_filename}")


def set_kubernetes_version(
    config, kubernetes_version, cloud_provider, grab_latest_version=True
):
    cloud_provider_dict = {
        "aws": {
            "full_name": "amazon_web_services",
            "k8s_version_checker_func": amazon_web_services.kubernetes_versions,
        },
        "azure": {
            "full_name": "azure",
            "k8s_version_checker_func": azure_cloud.kubernetes_versions,
        },
        "do": {
            "full_name": "digital_ocean",
            "k8s_version_checker_func": digital_ocean.kubernetes_versions,
        },
        "gcp": {
            "full_name": "google_cloud_platform",
            "k8s_version_checker_func": google_cloud.kubernetes_versions,
        },
    }
    cloud_full_name = cloud_provider_dict[cloud_provider]["full_name"]
    func = cloud_provider_dict[cloud_provider]["k8s_version_checker_func"]
    cloud_config = config[cloud_full_name]

    def _raise_value_error(cloud_provider, k8s_versions):
        raise ValueError(
            f"\nInvalid `kubernetes-version` provided: {kubernetes_version}.\nPlease select from one of the following {cloud_provider.upper()} supported Kubernetes versions: {k8s_versions} or omit flag to use latest Kubernetes version available."
        )

    def _check_and_set_kubernetes_version(
        kubernetes_version=kubernetes_version,
        cloud_provider=cloud_provider,
        cloud_config=cloud_config,
        func=func,
    ):
        region = cloud_config["region"]

        # to avoid using cloud provider SDK
        # set QHUB_K8S_VERSION environment variable
        if not QHUB_K8S_VERSION:
            k8s_versions = func(region)
        else:
            k8s_versions = [QHUB_K8S_VERSION]

        if kubernetes_version:
            if kubernetes_version in k8s_versions:
                cloud_config["kubernetes_version"] = kubernetes_version
            else:
                _raise_value_error(cloud_provider, k8s_versions)
        elif grab_latest_version:
            cloud_config["kubernetes_version"] = k8s_versions[-1]
        else:
            # grab oldest version
            cloud_config["kubernetes_version"] = k8s_versions[0]

    return _check_and_set_kubernetes_version()


@contextlib.contextmanager
def modified_environ(*remove: List[str], **update: Dict[str, str]):
    """
    https://stackoverflow.com/questions/2059482/python-temporarily-modify-the-current-processs-environment/51754362
    Temporarily updates the ``os.environ`` dictionary in-place.

    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations.

    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]


@contextlib.contextmanager
def kubernetes_provider_context(kubernetes_credentials: Dict[str, str]):
    credential_mapping = {
        "config_path": "KUBE_CONFIG_PATH",
        "config_context": "KUBE_CTX",
        "username": "KUBE_USER",
        "password": "KUBE_PASSWORD",
        "client_certificate": "KUBE_CLIENT_CERT_DATA",
        "client_key": "KUBE_CLIENT_KEY_DATA",
        "cluster_ca_certificate": "KUBE_CLUSTER_CA_CERT_DATA",
        "host": "KUBE_HOST",
        "token": "KUBE_TOKEN",
    }

    credentials = {
        credential_mapping[k]: v
        for k, v in kubernetes_credentials.items()
        if v is not None
    }
    with modified_environ(**credentials):
        yield


@contextlib.contextmanager
def keycloak_provider_context(keycloak_credentials: Dict[str, str]):
    credential_mapping = {
        "client_id": "KEYCLOAK_CLIENT_ID",
        "url": "KEYCLOAK_URL",
        "username": "KEYCLOAK_USER",
        "password": "KEYCLOAK_PASSWORD",
        "realm": "KEYCLOAK_REALM",
    }

    credentials = {credential_mapping[k]: v for k, v in keycloak_credentials.items()}
    with modified_environ(**credentials):
        yield


def deep_merge(*args):
    """Deep merge multiple dictionaries.

    >>> value_1 = {
    'a': [1, 2],
    'b': {'c': 1, 'z': [5, 6]},
    'e': {'f': {'g': {}}},
    'm': 1,
    }

    >>> value_2 = {
        'a': [3, 4],
        'b': {'d': 2, 'z': [7]},
        'e': {'f': {'h': 1}},
        'm': [1],
    }

    >>> print(deep_merge(value_1, value_2))
    {'m': 1, 'e': {'f': {'g': {}, 'h': 1}}, 'b': {'d': 2, 'c': 1, 'z': [5, 6, 7]}, 'a': [1, 2, 3,  4]}
    """
    if len(args) == 1:
        return args[0]
    elif len(args) > 2:
        return functools.reduce(deep_merge, args, {})
    else:  # length 2
        d1, d2 = args

    if isinstance(d1, dict) and isinstance(d2, dict):
        d3 = {}
        for key in d1.keys() | d2.keys():
            if key in d1 and key in d2:
                d3[key] = deep_merge(d1[key], d2[key])
            elif key in d1:
                d3[key] = d1[key]
            elif key in d2:
                d3[key] = d2[key]
        return d3
    elif isinstance(d1, list) and isinstance(d2, list):
        return [*d1, *d2]
    else:  # if they don't match use left one
        return d1


def set_docker_image_tag() -> str:
    """Set docker image tag for `jupyterlab`, `jupyterhub`, and `dask-worker`."""
    try:
        qhub_image_tag = get_latest_repo_tag(DOCKER_IMAGE_OWNER, DOCKER_IMAGE_REPO)
    except ConnectionError:
        print(
            f"Unable to connect to the GitHub API, falling back to the default value: {DEFAULT_QHUB_IMAGE_TAG}."
        )
        qhub_image_tag = DEFAULT_QHUB_IMAGE_TAG

    if QHUB_IMAGE_TAG:
        qhub_image_tag = QHUB_IMAGE_TAG

    return qhub_image_tag


def set_qhub_dask_version() -> str:
    """Set version of `qhub-dask` meta package."""
    try:
        resp = requests.get(CONDA_FORGE_CHANNEL_DATA_URL)
        qhub_dask_version = resp.json()["packages"]["qhub-dask"]["version"]
        resp.raise_for_status()
    except ConnectionError:
        print(
            f"Unable to connect to the Conda-Forge channel data, falling back to the default value: {DEFAULT_QHUB_DASK_VERSION}"
        )
        qhub_dask_version = DEFAULT_QHUB_DASK_VERSION

    if QHUB_DASK_VERSION:
        qhub_dask_version = QHUB_DASK_VERSION

    return qhub_dask_version



import pathlib

from qhub.provider.cicd.linter import comment_on_pr
from qhub.schema import verify
from qhub.utils import load_yaml


def create_validate_subcommand(subparser):
    subparser = subparser.add_parser("validate")
    subparser.add_argument(
        "configdeprecated",
        help="qhub configuration yaml file (deprecated - please pass in as -c/--config flag)",
        nargs="?",
    )
    subparser.add_argument(
        "-c", "--config", help="qhub configuration yaml file", required=False
    )
    subparser.add_argument(
        "--enable-commenting", help="Turn on PR commenting", action="store_true"
    )
    subparser.set_defaults(func=handle_validate)


def handle_validate(args):
    if args.configdeprecated and args.config:
        raise ValueError(
            "Please pass in -c/--config flag specifying your qhub-config.yaml file, and do NOT pass it as a standalone argument"
        )

    config_filename = args.config or args.configdeprecated
    if not config_filename:
        raise ValueError(
            "Please pass in a qhub-config.yaml filename using the -c/--config argument"
        )

    config_filename = pathlib.Path(args.config or args.configdeprecated)
    if not config_filename.is_file():
        raise ValueError(
            f"passed in configuration filename={config_filename} must exist"
        )

    config = load_yaml(config_filename)

    if args.enable_commenting:
        # for PR's only
        comment_on_pr(config)
    else:
        verify(config)



import json
import os
import pathlib
import textwrap

import requests

from qhub.schema import verify


def qhub_validate(config):
    # Gather the output of `qhub validate`.
    print("Validate: info: validating QHub configuration in qhub-config.yaml")

    def parse_validation(message):
        # this will just separate things for now, but can be enhanced
        return str(message)

    try:
        verify(config)
        msg = "validate: info: successfully validated QHub configuration"
        print(msg)
        return True, msg, 0

    except BaseException as e:
        msg = "validate: error: failed to validate QHub configuration."
        print(msg)
        validate_comment = parse_validation(e)
        validate_comment_wrapper = f"\n```\n{validate_comment}\n``` "
        return False, validate_comment_wrapper, 1


def generate_lint_message(config):

    # prep for linting
    pr_config = pathlib.Path("qhub-config.yaml")
    # lint/validate qhub-config.yaml
    all_pass, messages, validate_code = qhub_validate(config)

    pass_lint = textwrap.dedent(
        """
            This is an automatic response from the QHub linter.
            I just wanted to let you know that I linted your `qhub-config.yaml` in your PR and I didn't find any
            problems.
            """
    )

    # it should be better to parse this messages first
    bad_lint = (
        textwrap.dedent(
            """
            This is an automatic response from the QHub linter.
            I just wanted to let you know that I linted your `qhub-config.yaml` in your PR and found some errors:\n"""
        )
        + f"{messages}"
    )

    if not pr_config:
        status = "no configuration file"
        message = textwrap.dedent(
            """
            This is an automatic response from the QHub linter.
            I was trying to look for the `qhub-config.yaml` file to lint for you, but couldn't find any...
            """
        )

    elif all_pass:
        status = "Success"
        message = pass_lint
    else:
        status = "Failure"
        message = bad_lint

    lint = {
        "message": f"#### `qhub validate` {status} \n" + message,
        "code": validate_code,
    }
    return lint


def comment_on_pr(config):
    lint = generate_lint_message(config)
    message = lint["message"]
    exitcode = lint["code"]

    print(
        "If the comment was not published, the following would "
        "have been the message:\n{}".format(message)
    )

    # comment on PR
    owner, repo_name = os.environ["REPO_NAME"].split("/")
    pr_id = os.environ["PR_NUMBER"]

    token = os.environ["GITHUB_TOKEN"]
    url = f"https://api.github.com/repos/{owner}/{repo_name}/issues/{pr_id}/comments"

    payload = {"body": message}
    headers = {"Content-Type": "application/json", "Authorization": f"token {token}"}
    requests.post(url=url, headers=headers, data=json.dumps(payload))

    return exit(exitcode)



import json
import logging
import os
import tempfile

import requests
from conda_store_server import api, orm, schema
from conda_store_server.server.auth import GenericOAuthAuthentication
from conda_store_server.server.dependencies import get_conda_store
from conda_store_server.storage import S3Storage


def conda_store_config(path="/var/lib/conda-store/config.json"):
    with open(path) as f:
        return json.load(f)


config = conda_store_config()


# ==================================
#      conda-store settings
# ==================================
c.CondaStore.storage_class = S3Storage
c.CondaStore.store_directory = "/home/conda/"
c.CondaStore.database_url = f"postgresql+psycopg2://{config['postgres-username']}:{config['postgres-password']}@{config['postgres-service']}/conda-store"
c.CondaStore.redis_url = (
    f"redis://:{config['redis-password']}@{config['redis-service']}:6379/0"
)
c.CondaStore.default_uid = 1000
c.CondaStore.default_gid = 100
c.CondaStore.default_permissions = "775"
c.CondaStore.conda_included_packages = ["ipykernel"]

c.S3Storage.internal_endpoint = f"{config['minio-service']}:9000"
c.S3Storage.internal_secure = False
c.S3Storage.external_endpoint = f"{config['external-url']}:9080"
c.S3Storage.external_secure = True
c.S3Storage.access_key = config["minio-username"]
c.S3Storage.secret_key = config["minio-password"]
c.S3Storage.region = "us-east-1"  # minio region default
c.S3Storage.bucket_name = "conda-store"


# ==================================
#        server settings
# ==================================
c.CondaStoreServer.log_level = logging.INFO
c.CondaStoreServer.enable_ui = True
c.CondaStoreServer.enable_api = True
c.CondaStoreServer.enable_registry = True
c.CondaStoreServer.enable_metrics = True
c.CondaStoreServer.address = "0.0.0.0"
c.CondaStoreServer.port = 5000
c.CondaStoreServer.behind_proxy = True
# This MUST start with `/`
c.CondaStoreServer.url_prefix = "/conda-store"


# ==================================
#         auth settings
# ==================================

c.GenericOAuthAuthentication.access_token_url = config["openid-config"]["token_url"]
c.GenericOAuthAuthentication.authorize_url = config["openid-config"][
    "authentication_url"
]
c.GenericOAuthAuthentication.user_data_url = config["openid-config"]["userinfo_url"]
c.GenericOAuthAuthentication.oauth_callback_url = (
    f"https://{config['external-url']}/conda-store/oauth_callback"
)
c.GenericOAuthAuthentication.client_id = config["openid-config"]["client_id"]
c.GenericOAuthAuthentication.client_secret = config["openid-config"]["client_secret"]
c.GenericOAuthAuthentication.access_scope = "profile"
c.GenericOAuthAuthentication.user_data_key = "preferred_username"
c.GenericOAuthAuthentication.tls_verify = False


class KeyCloakAuthentication(GenericOAuthAuthentication):
    async def authenticate(self, request):
        # 1. using the callback_url code and state in request
        oauth_access_token = self._get_oauth_token(request)
        if oauth_access_token is None:
            return None  # authentication failed

        response = requests.get(
            self.user_data_url,
            headers={"Authorization": f"Bearer {oauth_access_token}"},
            verify=self.tls_verify,
        )
        response.raise_for_status()
        user_data = response.json()

        role_mappings = {
            "conda_store_admin": "admin",
            "conda_store_developer": "developer",
            "conda_store_viewer": "viewer",
        }
        roles = {
            role_mappings[role]
            for role in user_data.get("roles", [])
            if role in role_mappings
        }
        username = user_data["preferred_username"]
        namespaces = {username, "default", "filesystem"}
        role_bindings = {
            f"{username}/*": {"admin"},
            "filesystem/*": {"viewer"},
            "default/*": roles,
        }

        for group in user_data.get("groups", []):
            # Use only the base name of Keycloak groups
            group_name = os.path.basename(group)
            namespaces.add(group_name)
            role_bindings[f"{group_name}/*"] = roles

        conda_store = get_conda_store(request)
        for namespace in namespaces:
            _namespace = api.get_namespace(conda_store.db, name=namespace)
            if _namespace is None:
                conda_store.db.add(orm.Namespace(name=namespace))
                conda_store.db.commit()

        return schema.AuthenticationToken(
            primary_namespace=username,
            role_bindings=role_bindings,
        )


c.CondaStoreServer.authentication_class = KeyCloakAuthentication

# ==================================
#         worker settings
# ==================================
c.CondaStoreWorker.log_level = logging.INFO
c.CondaStoreWorker.watch_paths = ["/opt/environments"]
c.CondaStoreWorker.concurrency = 4

# extra-settings to apply simply as `c.Class.key = value`
conda_store_settings = config["extra-settings"]
for classname, attributes in conda_store_settings.items():
    for attribute, value in attributes.items():
        setattr(getattr(c, classname), attribute, value)

# run arbitrary python code
# compiling makes debugging easier: https://stackoverflow.com/a/437857
extra_config_filename = os.path.join(tempfile.gettempdir(), "extra-config.py")
extra_config = config.get("extra-config", "")
with open(extra_config_filename, "w") as f:
    f.write(extra_config)
exec(compile(source=extra_config, filename=extra_config_filename, mode="exec"))



import pathlib

from qhub.render import render_template
from qhub.schema import verify
from qhub.utils import QHUB_GH_BRANCH, load_yaml


def create_render_subcommand(subparser):
    subparser = subparser.add_parser("render")
    subparser.add_argument("-o", "--output", default="./", help="output directory")
    subparser.add_argument(
        "-c", "--config", help="qhub configuration yaml file", required=True
    )
    subparser.add_argument(
        "--dry-run",
        action="store_true",
        help="simulate rendering files without actually writing or updating any files",
    )
    subparser.set_defaults(func=handle_render)


def handle_render(args):

    if QHUB_GH_BRANCH:
        print(
            f"Modifying the version of `qhub` used by git-ops workflows, installing `qhub` from branch: {QHUB_GH_BRANCH}"
        )

    config_filename = pathlib.Path(args.config)
    if not config_filename.is_file():
        raise ValueError(
            f"passed in configuration filename={config_filename} must exist"
        )

    config = load_yaml(config_filename)

    verify(config)

    render_template(args.output, args.config, force=True, dry_run=args.dry_run)



import re
import uuid

import paramiko
import pytest

from tests_deployment import constants
from tests_deployment.utils import (
    escape_string,
    get_jupyterhub_token,
    monkeypatch_ssl_context,
)

monkeypatch_ssl_context()

TIMEOUT_SECS = 300


@pytest.fixture
def paramiko_object():
    """Connects to JupyterHub ssh cluster from outside the cluster."""
    api_token = get_jupyterhub_token("jupyterhub-ssh")

    try:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
        client.connect(
            hostname=constants.QHUB_HOSTNAME,
            port=8022,
            username=constants.KEYCLOAK_USERNAME,
            password=api_token,
            # wait 5 minutes for jupyterlab server/terminal to spin up
            auth_timeout=5 * 60,
        )
        yield client
    finally:
        client.close()


def run_command(command, stdin, stdout, stderr):
    delimiter = uuid.uuid4().hex
    stdin.write(f"echo {delimiter}start; {command}; echo {delimiter}end\n")

    output = []

    line = stdout.readline()
    while not re.match(f"^{delimiter}start$", line.strip()):
        line = stdout.readline()

    line = stdout.readline()
    if delimiter not in line:
        output.append(line)

    while not re.match(f"^{delimiter}end$", line.strip()):
        line = stdout.readline()
        if delimiter not in line:
            output.append(line)

    return "".join(output).strip()


@pytest.mark.timeout(TIMEOUT_SECS)
def test_simple_jupyterhub_ssh(paramiko_object):
    stdin, stdout, stderr = paramiko_object.exec_command("")


@pytest.mark.timeout(TIMEOUT_SECS)
def test_print_jupyterhub_ssh(paramiko_object):
    stdin, stdout, stderr = paramiko_object.exec_command("")

    # commands to run and just print the output
    commands_print = [
        "id",
        "env",
        "conda info",
        "df -h",
        "ls -la",
        "umask",
    ]

    for command in commands_print:
        print(f'COMMAND: "{command}"')
        print(run_command(command, stdin, stdout, stderr))


@pytest.mark.timeout(TIMEOUT_SECS)
def test_exact_jupyterhub_ssh(paramiko_object):
    stdin, stdout, stderr = paramiko_object.exec_command("")

    # commands to run and exactly match output
    commands_exact = [
        ("id -u", "1000"),
        ("id -g", "100"),
        ("whoami", constants.KEYCLOAK_USERNAME),
        ("pwd", f"/home/{constants.KEYCLOAK_USERNAME}"),
        ("echo $HOME", f"/home/{constants.KEYCLOAK_USERNAME}"),
        ("conda activate default && echo $CONDA_PREFIX", "/opt/conda/envs/default"),
        ("hostname", f"jupyter-{escape_string(constants.KEYCLOAK_USERNAME)}"),
    ]

    for command, output in commands_exact:
        assert output == run_command(command, stdin, stdout, stderr)


@pytest.mark.timeout(TIMEOUT_SECS)
def test_contains_jupyterhub_ssh(paramiko_object):
    stdin, stdout, stderr = paramiko_object.exec_command("")

    # commands to run and string need to be contained in output
    commands_contain = [
        ("ls -la", ".bashrc"),
        ("ls -la /shared/examples", "dashboard_panel.ipynb"),
        ("cat ~/.bashrc", "Managed by QHub"),
        ("cat ~/.profile", "Managed by QHub"),
        ("cat ~/.bash_logout", "Managed by QHub"),
    ]

    for command, output in commands_contain:
        assert output in run_command(command, stdin, stdout, stderr)



from typing import Dict

from qhub.provider.terraform import Data, Provider, TerraformBackend, tf_render_objects
from qhub.utils import deep_merge


def QHubAWSProvider(qhub_config: Dict):
    return Provider("aws", region=qhub_config["amazon_web_services"]["region"])


def QHubGCPProvider(qhub_config: Dict):
    return Provider(
        "google",
        project=qhub_config["google_cloud_platform"]["project"],
        region=qhub_config["google_cloud_platform"]["region"],
    )


def QHubAzureProvider(qhub_config: Dict):
    return Provider("azurerm", features={})


def QHubDigitalOceanProvider(qhub_config: Dict):
    return Provider("digitalocean")


def QHubKubernetesProvider(qhub_config: Dict):
    if qhub_config["provider"] == "aws":
        cluster_name = f"{qhub_config['project_name']}-{qhub_config['namespace']}"
        # The AWS provider needs to be added, as we are using aws related resources #1254
        return deep_merge(
            Data("aws_eks_cluster", "default", name=cluster_name),
            Data("aws_eks_cluster_auth", "default", name=cluster_name),
            Provider("aws", region=qhub_config["amazon_web_services"]["region"]),
            Provider(
                "kubernetes",
                experiments={"manifest_resource": True},
                host="${data.aws_eks_cluster.default.endpoint}",
                cluster_ca_certificate="${base64decode(data.aws_eks_cluster.default.certificate_authority[0].data)}",
                token="${data.aws_eks_cluster_auth.default.token}",
            ),
        )
    return Provider(
        "kubernetes",
        experiments={"manifest_resource": True},
    )


def QHubHelmProvider(qhub_config: Dict):
    if qhub_config["provider"] == "aws":
        cluster_name = f"{qhub_config['project_name']}-{qhub_config['namespace']}"

        return deep_merge(
            Data("aws_eks_cluster", "default", name=cluster_name),
            Data("aws_eks_cluster_auth", "default", name=cluster_name),
            Provider(
                "helm",
                kubernetes=dict(
                    host="${data.aws_eks_cluster.default.endpoint}",
                    cluster_ca_certificate="${base64decode(data.aws_eks_cluster.default.certificate_authority[0].data)}",
                    token="${data.aws_eks_cluster_auth.default.token}",
                ),
            ),
        )
    return Provider("helm")


def QHubTerraformState(directory: str, qhub_config: Dict):
    if qhub_config["terraform_state"]["type"] == "local":
        return {}
    elif qhub_config["terraform_state"]["type"] == "existing":
        return TerraformBackend(
            qhub_config["terraform_state"]["backend"],
            **qhub_config["terraform_state"]["config"],
        )
    elif qhub_config["provider"] == "aws":
        return TerraformBackend(
            "s3",
            bucket=f"{qhub_config['project_name']}-{qhub_config['namespace']}-terraform-state",
            key=f"terraform/{qhub_config['project_name']}-{qhub_config['namespace']}/{directory}.tfstate",
            region=qhub_config["amazon_web_services"]["region"],
            encrypt=True,
            dynamodb_table=f"{qhub_config['project_name']}-{qhub_config['namespace']}-terraform-state-lock",
        )
    elif qhub_config["provider"] == "gcp":
        return TerraformBackend(
            "gcs",
            bucket=f"{qhub_config['project_name']}-{qhub_config['namespace']}-terraform-state",
            prefix=f"terraform/{qhub_config['project_name']}/{directory}",
        )
    elif qhub_config["provider"] == "do":
        return TerraformBackend(
            "s3",
            endpoint=f"{qhub_config['digital_ocean']['region']}.digitaloceanspaces.com",
            region="us-west-1",  # fake aws region required by terraform
            bucket=f"{qhub_config['project_name']}-{qhub_config['namespace']}-terraform-state",
            key=f"terraform/{qhub_config['project_name']}-{qhub_config['namespace']}/{directory}.tfstate",
            skip_credentials_validation=True,
            skip_metadata_api_check=True,
        )
    elif qhub_config["provider"] == "azure":
        return TerraformBackend(
            "azurerm",
            resource_group_name=f"{qhub_config['project_name']}-{qhub_config['namespace']}-state",
            # storage account must be globally unique
            storage_account_name=f"{qhub_config['project_name']}{qhub_config['namespace']}{qhub_config['azure']['storage_account_postfix']}",
            container_name=f"{qhub_config['project_name']}-{qhub_config['namespace']}-state",
            key=f"terraform/{qhub_config['project_name']}-{qhub_config['namespace']}/{directory}",
        )
    elif qhub_config["provider"] == "existing":
        optional_kwargs = {}
        if "kube_context" in qhub_config["existing"]:
            optional_kwargs["confix_context"] = qhub_config["existing"]["kube_context"]
        return TerraformBackend(
            "kubernetes",
            secret_suffix=f"{qhub_config['project_name']}-{qhub_config['namespace']}-{directory}",
            load_config_file=True,
            **optional_kwargs,
        )
    elif qhub_config["provider"] == "local":
        optional_kwargs = {}
        if "kube_context" in qhub_config["local"]:
            optional_kwargs["confix_context"] = qhub_config["local"]["kube_context"]
        return TerraformBackend(
            "kubernetes",
            secret_suffix=f"{qhub_config['project_name']}-{qhub_config['namespace']}-{directory}",
            load_config_file=True,
            **optional_kwargs,
        )
    else:
        raise NotImplementedError("state not implemented")


def stage_01_terraform_state(config):
    if config["provider"] == "gcp":
        return {
            "stages/01-terraform-state/gcp/_qhub.tf.json": tf_render_objects(
                [
                    QHubGCPProvider(config),
                ]
            )
        }
    elif config["provider"] == "aws":
        return {
            "stages/01-terraform-state/aws/_qhub.tf.json": tf_render_objects(
                [
                    QHubAWSProvider(config),
                ]
            )
        }
    else:
        return {}


def stage_02_infrastructure(config):
    if config["provider"] == "gcp":
        return {
            "stages/02-infrastructure/gcp/_qhub.tf.json": tf_render_objects(
                [
                    QHubGCPProvider(config),
                    QHubTerraformState("02-infrastructure", config),
                ]
            )
        }
    elif config["provider"] == "do":
        return {
            "stages/02-infrastructure/do/_qhub.tf.json": tf_render_objects(
                [
                    QHubTerraformState("02-infrastructure", config),
                ]
            )
        }
    elif config["provider"] == "azure":
        return {
            "stages/02-infrastructure/azure/_qhub.tf.json": tf_render_objects(
                [
                    QHubTerraformState("02-infrastructure", config),
                ]
            ),
        }
    elif config["provider"] == "aws":
        return {
            "stages/02-infrastructure/aws/_qhub.tf.json": tf_render_objects(
                [
                    QHubAWSProvider(config),
                    QHubTerraformState("02-infrastructure", config),
                ]
            )
        }
    else:
        return {}


def stage_03_kubernetes_initialize(config):
    return {
        "stages/03-kubernetes-initialize/_qhub.tf.json": tf_render_objects(
            [
                QHubTerraformState("03-kubernetes-initialize", config),
                QHubKubernetesProvider(config),
                QHubHelmProvider(config),
            ]
        ),
    }


def stage_04_kubernetes_ingress(config):
    return {
        "stages/04-kubernetes-ingress/_qhub.tf.json": tf_render_objects(
            [
                QHubTerraformState("04-kubernetes-ingress", config),
                QHubKubernetesProvider(config),
                QHubHelmProvider(config),
            ]
        ),
    }


def stage_05_kubernetes_keycloak(config):
    return {
        "stages/05-kubernetes-keycloak/_qhub.tf.json": tf_render_objects(
            [
                QHubTerraformState("05-kubernetes-keycloak", config),
                QHubKubernetesProvider(config),
                QHubHelmProvider(config),
            ]
        ),
    }


def stage_06_kubernetes_keycloak_configuration(config):
    return {
        "stages/06-kubernetes-keycloak-configuration/_qhub.tf.json": tf_render_objects(
            [
                QHubTerraformState("06-kubernetes-keycloak-configuration", config),
            ]
        ),
    }


def stage_07_kubernetes_services(config):
    return {
        "stages/07-kubernetes-services/_qhub.tf.json": tf_render_objects(
            [
                QHubTerraformState("07-kubernetes-services", config),
                QHubKubernetesProvider(config),
                QHubHelmProvider(config),
            ]
        ),
    }


def stage_08_qhub_tf_extensions(config):
    return {
        "stages/08-qhub-tf-extensions/_qhub.tf.json": tf_render_objects(
            [
                QHubTerraformState("08-qhub-tf-extensions", config),
                QHubKubernetesProvider(config),
                QHubHelmProvider(config),
            ]
        ),
    }



import os


def stage_01_terraform_state(stage_outputs, config):
    if config["provider"] == "do":
        return [
            (
                "module.terraform-state.module.spaces.digitalocean_spaces_bucket.main",
                f"{config['digital_ocean']['region']},{config['project_name']}-{config['namespace']}-terraform-state",
            )
        ]
    elif config["provider"] == "gcp":
        return [
            (
                "module.terraform-state.module.gcs.google_storage_bucket.static-site",
                f"{config['project_name']}-{config['namespace']}-terraform-state",
            )
        ]
    elif config["provider"] == "azure":
        subscription_id = os.environ["ARM_SUBSCRIPTION_ID"]
        resource_name_prefix = f"{config['project_name']}-{config['namespace']}"
        state_resource_group_name = f"{resource_name_prefix}-state"
        state_resource_name_prefix_safe = resource_name_prefix.replace("-", "")
        resource_group_url = f"/subscriptions/{subscription_id}/resourceGroups/{state_resource_group_name}"

        return [
            (
                "module.terraform-state.azurerm_resource_group.terraform-state-resource-group",
                resource_group_url,
            ),
            (
                "module.terraform-state.azurerm_storage_account.terraform-state-storage-account",
                f"{resource_group_url}/providers/Microsoft.Storage/storageAccounts/{state_resource_name_prefix_safe}{config['azure']['storage_account_postfix']}",
            ),
            (
                "module.terraform-state.azurerm_storage_container.storage_container",
                f"https://{state_resource_name_prefix_safe}{config['azure']['storage_account_postfix']}.blob.core.windows.net/{resource_name_prefix}-state",
            ),
        ]
    elif config["provider"] == "aws":
        return [
            (
                "module.terraform-state.aws_s3_bucket.terraform-state",
                f"{config['project_name']}-{config['namespace']}-terraform-state",
            ),
            (
                "module.terraform-state.aws_dynamodb_table.terraform-state-lock",
                f"{config['project_name']}-{config['namespace']}-terraform-state-lock",
            ),
        ]



import json
import logging
import os
import re
import subprocess

from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table

logger = logging.getLogger(__name__)

INFRACOST_NOTE = """
QHub rely upon node-pools which is a usage resource but doesn't get captured in the above report. A general node-pool
will always have one node running will add quite an additional charge. Please check in with your cloud provider to see
the associated costs with node pools.

- Total Monthly Cost: The total monthly cost of the deployment of supported resources.
- Total Hourly Cost: The total hourly cost of the deployment of supported resources.
- Total Detected Costs: The total number of resources detected by Infracost.
- Total Supported Resources: The total number of resources supported by Infracost.
- Total Un-Supported Resources: The total number of resources unsupported by Infracost.
- Total Non-Priced Resources: The total number of resources that are not priced.
- Total Usage-Priced Resources: The total number of resources that are priced based on usage.
"""


def _check_infracost():
    """
    Check if infracost is installed
    """
    try:
        subprocess.check_output(["infracost", "--version"])
        return True
    except subprocess.CalledProcessError:
        return False


def _check_infracost_api_key():
    """
    Check if infracost API key is configured
    """
    try:
        subprocess.check_output(["infracost", "configure", "get", "api_key"])
        return True
    except subprocess.CalledProcessError:
        return False


def _set_currency_code(currency_code):
    """
    Specify the currency code for infracost
    """
    try:
        subprocess.check_output(
            ["infracost", "configure", "set", "currency", currency_code]
        )
        return True
    except subprocess.CalledProcessError:
        return False


def _run_infracost(path):
    """
    Run infracost on the given path and return the JSON output
    """
    try:
        process = subprocess.Popen(
            ["infracost", "breakdown", "--path", path, "--format", "json"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        infracost_data = json.loads(
            re.search("({.+})", stdout.decode("UTF-8"))
            .group(0)
            .replace("u'", '"')
            .replace("'", '"')
        )
        return infracost_data
    except subprocess.CalledProcessError:
        return None
    except AttributeError:
        return None


def _enable_infracost_dashboard():
    """
    Enable infracost dashboard
    """
    try:
        subprocess.check_output(
            ["infracost", "configure", "set", "enable_dashboard", "true"]
        )
        return True
    except subprocess.CalledProcessError:
        return False


def _disable_infracost_dashboard():
    """
    Disable infracost dashboard
    """
    try:
        subprocess.check_output(
            ["infracost", "configure", "set", "enable_dashboard", "false"]
        )
        return True
    except subprocess.CalledProcessError:
        return False


def infracost_diff(path, compare_to_path):
    """
    Compare the infracost report of the given path to a previous infracost report
    """
    try:
        process = subprocess.Popen(
            [
                "infracost",
                "diff",
                "--path",
                path,
                "--compare-to",
                compare_to_path,
                "--format",
                "json",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        diff_data = json.loads(
            re.search("({.+})", stdout.decode("UTF-8"))
            .group(0)
            .replace("u'", '"')
            .replace("'", '"')
        )
        return diff_data
    except subprocess.CalledProcessError:
        return None


def infracost_report(path, dashboard, file, currency_code, compare):
    """
    Generate a report of the infracost cost of the given path
    args:
        path: path to the qhub stages directory
    """
    console = Console()
    # If path is not provided, use the current directory with `stages` subdirectory
    if not path:
        path = os.path.join(os.getcwd(), "stages")

    # Checks if infracost is installed and an API key is configured
    if _check_infracost() and _check_infracost_api_key():
        _set_currency_code(currency_code)
        if not dashboard:
            _disable_infracost_dashboard()
        else:
            _enable_infracost_dashboard()

        # Check if the deployment is available on the given path
        if not os.path.exists(path):
            logger.error("Deployment is not available")
        else:
            data = _run_infracost(path)

            # If a user has asked for a JSON file, download it
            if file:
                # Convert data to JSON and save it to the given file
                with open(file, "w") as f:
                    json.dump(data, f)

            if data:
                cost_table = Table(title="Cost Breakdown")
                cost_table.add_column(
                    "Name", justify="right", style="cyan", no_wrap=True
                )
                cost_table.add_column(
                    "Cost", justify="right", style="cyan", no_wrap=True
                )

                cost_table.add_row("Total Monthly Cost", data["totalMonthlyCost"])
                cost_table.add_row("Total Hourly Cost", data["totalHourlyCost"])

                resource_table = Table(title="Resource Breakdown")
                resource_table.add_column(
                    "Name", justify="right", style="cyan", no_wrap=True
                )
                resource_table.add_column(
                    "Number", justify="right", style="cyan", no_wrap=True
                )

                resource_table.add_row(
                    "Total Detected Costs",
                    str(data["summary"]["totalDetectedResources"]),
                )
                resource_table.add_row(
                    "Total Supported Resources",
                    str(data["summary"]["totalSupportedResources"]),
                )
                resource_table.add_row(
                    "Total Un-Supported Resources",
                    str(data["summary"]["totalUnsupportedResources"]),
                )
                resource_table.add_row(
                    "Total Non-Priced Resources",
                    str(data["summary"]["totalNoPriceResources"]),
                )
                resource_table.add_row(
                    "Total Usage-Priced Resources",
                    str(data["summary"]["totalUsageBasedResources"]),
                )
                console.print(cost_table)
                console.print(resource_table)
                # If dashboard is enabled, display the dashboard
                if dashboard:
                    console.print(
                        f"Access the cost & resource breakdown dashboard here: {data['shareUrl']}\n"
                    )

                # If the user is comparing the cost of the current deployment to a previous deployment
                if compare:
                    diff_data = infracost_diff(path, compare)
                    diff_table = Table(title="Diff Breakdown")
                    diff_table.add_column(
                        "Name", justify="right", style="cyan", no_wrap=True
                    )
                    diff_table.add_column(
                        "Cost", justify="right", style="cyan", no_wrap=True
                    )
                    diff_table.add_row(
                        "Diff Monthly Cost", diff_data["diffTotalHourlyCost"]
                    )
                    diff_table.add_row(
                        "Diff Hourly Cost", diff_data["diffTotalMonthlyCost"]
                    )
                    console.print(diff_table)
                    # If dashboard is enabled, display the diff breakdown dashboard
                    if dashboard:
                        console.print(
                            f"Access the diff breakdown dashboard here: {diff_data['shareUrl']}\n"
                        )

                console.print(Markdown(INFRACOST_NOTE))
            else:
                logger.error(
                    "No data was generated. Please check your QHub configuration and generated stages."
                )
    else:
        logger.error("Infracost is not installed or the API key is not configured")



import logging
import os
import random
import re
import secrets
import string
import tempfile

import requests

from qhub.provider import git
from qhub.provider.cicd import github
from qhub.provider.oauth.auth0 import create_client
from qhub.utils import (
    check_cloud_credentials,
    namestr_regex,
    set_docker_image_tag,
    set_kubernetes_version,
    set_qhub_dask_version,
)

from .version import __version__

logger = logging.getLogger(__name__)

qhub_image_tag = set_docker_image_tag()
qhub_dask_version = set_qhub_dask_version()


BASE_CONFIGURATION = {
    "project_name": None,
    "provider": None,
    "domain": None,
    "certificate": {
        "type": "self-signed",
    },
    "security": {
        "authentication": None,
    },
    "default_images": {
        "jupyterhub": f"quay.io/nebari/nebari-jupyterhub:{qhub_image_tag}",
        "jupyterlab": f"quay.io/nebari/nebari-jupyterlab:{qhub_image_tag}",
        "dask_worker": f"quay.io/nebari/nebari-dask-worker:{qhub_image_tag}",
    },
    "storage": {"conda_store": "200Gi", "shared_filesystem": "200Gi"},
    "theme": {
        "jupyterhub": {
            "hub_title": None,
            "hub_subtitle": None,
            "welcome": None,
            "logo": "/hub/custom/images/jupyter_qhub_logo.svg",
            "primary_color": "#4f4173",
            "secondary_color": "#957da6",
            "accent_color": "#32C574",
            "text_color": "#111111",
            "h1_color": "#652e8e",
            "h2_color": "#652e8e",
            "version": f"v{__version__}",
        }
    },
    "helm_extensions": [],
    "monitoring": {
        "enabled": True,
    },
    "argo_workflows": {
        "enabled": True,
    },
    "kbatch": {
        "enabled": True,
    },
    "cdsdashboards": {
        "enabled": True,
        "cds_hide_user_named_servers": True,
        "cds_hide_user_dashboard_servers": False,
    },
}

CICD_CONFIGURATION = {
    "type": "PLACEHOLDER",
    "branch": "main",
    "commit_render": True,
}

AUTH_PASSWORD = {
    "type": "password",
}

AUTH_OAUTH_GITHUB = {
    "type": "GitHub",
    "config": {
        "client_id": "PLACEHOLDER",
        "client_secret": "PLACEHOLDER",
    },
}

AUTH_OAUTH_AUTH0 = {
    "type": "Auth0",
    "config": {
        "client_id": "PLACEHOLDER",
        "client_secret": "PLACEHOLDER",
        "auth0_subdomain": "PLACEHOLDER",
    },
}

LOCAL = {
    "node_selectors": {
        "general": {
            "key": "kubernetes.io/os",
            "value": "linux",
        },
        "user": {
            "key": "kubernetes.io/os",
            "value": "linux",
        },
        "worker": {
            "key": "kubernetes.io/os",
            "value": "linux",
        },
    }
}

EXISTING = {
    "node_selectors": {
        "general": {
            "key": "kubernetes.io/os",
            "value": "linux",
        },
        "user": {
            "key": "kubernetes.io/os",
            "value": "linux",
        },
        "worker": {
            "key": "kubernetes.io/os",
            "value": "linux",
        },
    }
}

DIGITAL_OCEAN = {
    "region": "nyc3",
    "kubernetes_version": "PLACEHOLDER",
    "node_groups": {
        "general": {"instance": "g-8vcpu-32gb", "min_nodes": 1, "max_nodes": 1},
        "user": {"instance": "g-4vcpu-16gb", "min_nodes": 1, "max_nodes": 5},
        "worker": {"instance": "g-4vcpu-16gb", "min_nodes": 1, "max_nodes": 5},
    },
}
# Digital Ocean image slugs are listed here https://slugs.do-api.dev/

GOOGLE_PLATFORM = {
    "project": "PLACEHOLDER",
    "region": "us-central1",
    "kubernetes_version": "PLACEHOLDER",
    "node_groups": {
        "general": {"instance": "n1-standard-8", "min_nodes": 1, "max_nodes": 1},
        "user": {"instance": "n1-standard-4", "min_nodes": 0, "max_nodes": 5},
        "worker": {"instance": "n1-standard-4", "min_nodes": 0, "max_nodes": 5},
    },
}

AZURE = {
    "region": "Central US",
    "kubernetes_version": "PLACEHOLDER",
    "node_groups": {
        "general": {
            "instance": "Standard_D8_v3",
            "min_nodes": 1,
            "max_nodes": 1,
        },
        "user": {"instance": "Standard_D4_v3", "min_nodes": 0, "max_nodes": 5},
        "worker": {
            "instance": "Standard_D4_v3",
            "min_nodes": 0,
            "max_nodes": 5,
        },
    },
    "storage_account_postfix": "".join(
        random.choices("abcdefghijklmnopqrstuvwxyz0123456789", k=8)
    ),
}

AMAZON_WEB_SERVICES = {
    "region": "us-west-2",
    "kubernetes_version": "PLACEHOLDER",
    "node_groups": {
        "general": {"instance": "m5.2xlarge", "min_nodes": 1, "max_nodes": 1},
        "user": {"instance": "m5.xlarge", "min_nodes": 1, "max_nodes": 5},
        "worker": {"instance": "m5.xlarge", "min_nodes": 1, "max_nodes": 5},
    },
}

DEFAULT_PROFILES = {
    "jupyterlab": [
        {
            "display_name": "Small Instance",
            "description": "Stable environment with 2 cpu / 8 GB ram",
            "default": True,
            "kubespawner_override": {
                "cpu_limit": 2,
                "cpu_guarantee": 1.5,
                "mem_limit": "8G",
                "mem_guarantee": "5G",
            },
        },
        {
            "display_name": "Medium Instance",
            "description": "Stable environment with 4 cpu / 16 GB ram",
            "kubespawner_override": {
                "cpu_limit": 4,
                "cpu_guarantee": 3,
                "mem_limit": "16G",
                "mem_guarantee": "10G",
            },
        },
    ],
    "dask_worker": {
        "Small Worker": {
            "worker_cores_limit": 2,
            "worker_cores": 1.5,
            "worker_memory_limit": "8G",
            "worker_memory": "5G",
            "worker_threads": 2,
        },
        "Medium Worker": {
            "worker_cores_limit": 4,
            "worker_cores": 3,
            "worker_memory_limit": "16G",
            "worker_memory": "10G",
            "worker_threads": 4,
        },
    },
}

DEFAULT_ENVIRONMENTS = {
    "environment-dask.yaml": {
        "name": "dask",
        "channels": ["conda-forge"],
        "dependencies": [
            "python",
            "ipykernel",
            "ipywidgets",
            f"qhub-dask =={qhub_dask_version}",
            "python-graphviz",
            "pyarrow",
            "s3fs",
            "gcsfs",
            "numpy",
            "numba",
            "pandas",
            {
                "pip": [
                    "kbatch",
                ],
            },
        ],
    },
    "environment-dashboard.yaml": {
        "name": "dashboard",
        "channels": ["conda-forge"],
        "dependencies": [
            "python==3.9.13",
            "ipykernel==6.15.1",
            "ipywidgets==7.7.1",
            f"qhub-dask=={qhub_dask_version}",
            "param==1.12.2",
            "python-graphviz==0.20.1",
            "matplotlib==3.3.2",
            "panel==0.13.1",
            "voila==0.3.6",
            "streamlit==1.10.0",
            "dash==2.6.1",
            "cdsdashboards-singleuser==0.6.2",
        ],
    },
}


def render_config(
    project_name,
    qhub_domain,
    cloud_provider,
    ci_provider,
    repository,
    auth_provider,
    namespace=None,
    repository_auto_provision=False,
    auth_auto_provision=False,
    terraform_state=None,
    kubernetes_version=None,
    disable_prompt=False,
    ssl_cert_email=None,
):
    config = BASE_CONFIGURATION.copy()
    config["provider"] = cloud_provider

    if ci_provider is not None and ci_provider != "none":
        config["ci_cd"] = CICD_CONFIGURATION.copy()
        config["ci_cd"]["type"] = ci_provider

    if terraform_state is not None:
        config["terraform_state"] = {"type": terraform_state}

    if project_name is None and not disable_prompt:
        project_name = input("Provide project name: ")
    config["project_name"] = project_name

    if not re.match(namestr_regex, project_name):
        raise ValueError(
            "project name should contain only letters and hyphens/underscores (but not at the start or end)"
        )

    if namespace is not None:
        config["namespace"] = namespace

    if not re.match(namestr_regex, namespace):
        raise ValueError(
            "namespace should contain only letters and hyphens/underscores (but not at the start or end)"
        )

    if qhub_domain is None and not disable_prompt:
        qhub_domain = input("Provide domain: ")
    config["domain"] = qhub_domain

    # In qhub_version only use major.minor.patch version - drop any pre/post/dev suffixes
    config["qhub_version"] = __version__

    # Generate default password for Keycloak root user and also example-user if using password auth
    default_password = "".join(
        secrets.choice(string.ascii_letters + string.digits) for i in range(16)
    )

    # Save default password to file
    default_password_filename = os.path.join(
        tempfile.gettempdir(), "QHUB_DEFAULT_PASSWORD"
    )
    with open(default_password_filename, "w") as f:
        f.write(default_password)
    os.chmod(default_password_filename, 0o700)

    print(
        f"Securely generated default random password={default_password} for Keycloak root user stored at path={default_password_filename}"
    )

    config["theme"]["jupyterhub"]["hub_title"] = f"QHub - { project_name }"
    config["theme"]["jupyterhub"][
        "welcome"
    ] = f"""Welcome to { qhub_domain }. It is maintained by <a href="http://quansight.com">Quansight staff</a>. The hub's configuration is stored in a github repository based on <a href="https://github.com/Quansight/qhub/">https://github.com/Quansight/qhub/</a>. To provide feedback and report any technical problems, please use the <a href="https://github.com/Quansight/qhub/issues">github issue tracker</a>."""

    if auth_provider == "github":
        config["security"]["authentication"] = AUTH_OAUTH_GITHUB.copy()
        print(
            "Visit https://github.com/settings/developers and create oauth application"
        )
        print(f"  set the homepage to: https://{qhub_domain}/")
        print(
            f"  set the callback_url to: https://{qhub_domain}/auth/realms/qhub/broker/github/endpoint"
        )

        if not disable_prompt:
            config["security"]["authentication"]["config"]["client_id"] = input(
                "Github client_id: "
            )
            config["security"]["authentication"]["config"]["client_secret"] = input(
                "Github client_secret: "
            )
    elif auth_provider == "auth0":
        config["security"]["authentication"] = AUTH_OAUTH_AUTH0.copy()

    elif auth_provider == "password":
        config["security"]["authentication"] = AUTH_PASSWORD.copy()

    # Always use default password for keycloak root
    config["security"].setdefault("keycloak", {})[
        "initial_root_password"
    ] = default_password

    if cloud_provider == "do":
        config["theme"]["jupyterhub"][
            "hub_subtitle"
        ] = "Autoscaling Compute Environment on Digital Ocean"
        config["digital_ocean"] = DIGITAL_OCEAN.copy()
        set_kubernetes_version(config, kubernetes_version, cloud_provider)

    elif cloud_provider == "gcp":
        config["theme"]["jupyterhub"][
            "hub_subtitle"
        ] = "Autoscaling Compute Environment on Google Cloud Platform"
        config["google_cloud_platform"] = GOOGLE_PLATFORM.copy()
        set_kubernetes_version(config, kubernetes_version, cloud_provider)

        if "PROJECT_ID" in os.environ:
            config["google_cloud_platform"]["project"] = os.environ["PROJECT_ID"]
        elif not disable_prompt:
            config["google_cloud_platform"]["project"] = input(
                "Enter Google Cloud Platform Project ID: "
            )

    elif cloud_provider == "azure":
        config["theme"]["jupyterhub"][
            "hub_subtitle"
        ] = "Autoscaling Compute Environment on Azure"
        config["azure"] = AZURE.copy()
        set_kubernetes_version(config, kubernetes_version, cloud_provider)

    elif cloud_provider == "aws":
        config["theme"]["jupyterhub"][
            "hub_subtitle"
        ] = "Autoscaling Compute Environment on Amazon Web Services"
        config["amazon_web_services"] = AMAZON_WEB_SERVICES.copy()
        set_kubernetes_version(config, kubernetes_version, cloud_provider)
        if "AWS_DEFAULT_REGION" in os.environ:
            config["amazon_web_services"]["region"] = os.environ["AWS_DEFAULT_REGION"]

    elif cloud_provider == "existing":
        config["theme"]["jupyterhub"][
            "hub_subtitle"
        ] = "Autoscaling Compute Environment"
        config["existing"] = EXISTING.copy()

    elif cloud_provider == "local":
        config["theme"]["jupyterhub"][
            "hub_subtitle"
        ] = "Autoscaling Compute Environment"
        config["local"] = LOCAL.copy()

    config["profiles"] = DEFAULT_PROFILES.copy()
    config["environments"] = DEFAULT_ENVIRONMENTS.copy()

    if ssl_cert_email is not None:
        if not re.match("^[^ @]+@[^ @]+\\.[^ @]+$", ssl_cert_email):
            raise ValueError("ssl-cert-email should be a valid email address")
        config["certificate"] = {
            "type": "lets-encrypt",
            "acme_email": ssl_cert_email,
            "acme_server": "https://acme-v02.api.letsencrypt.org/directory",
        }

    if auth_auto_provision:
        if auth_provider == "auth0":
            auth0_auto_provision(config)

    if repository_auto_provision:
        GITHUB_REGEX = "(https://)?github.com/([^/]+)/([^/]+)/?"
        if re.search(GITHUB_REGEX, repository):
            match = re.search(GITHUB_REGEX, repository)
            git_repository = github_auto_provision(
                config, match.group(2), match.group(3)
            )
            git_repository_initialize(git_repository)
        else:
            raise ValueError(
                f"Repository to be auto-provisioned is not the full URL of a GitHub repo: {repository}"
            )

    return config


def github_auto_provision(config, owner, repo):
    check_cloud_credentials(
        config
    )  # We may need env vars such as AWS_ACCESS_KEY_ID depending on provider

    already_exists = True
    try:
        github.get_repository(owner, repo)
    except requests.exceptions.HTTPError:
        # repo not found
        already_exists = False

    if not already_exists:
        try:
            github.create_repository(
                owner,
                repo,
                description=f'QHub {config["project_name"]}-{config["provider"]}',
                homepage=f'https://{config["domain"]}',
            )
        except requests.exceptions.HTTPError as he:
            raise ValueError(
                f"Unable to create GitHub repo https://github.com/{owner}/{repo} - error message from GitHub is: {he}"
            )
    else:
        logger.warn(f"GitHub repo https://github.com/{owner}/{repo} already exists")

    try:
        # Secrets
        if config["provider"] == "do":
            for name in {
                "AWS_ACCESS_KEY_ID",
                "AWS_SECRET_ACCESS_KEY",
                "SPACES_ACCESS_KEY_ID",
                "SPACES_SECRET_ACCESS_KEY",
                "DIGITALOCEAN_TOKEN",
            }:
                github.update_secret(owner, repo, name, os.environ[name])
        elif config["provider"] == "aws":
            for name in {
                "AWS_ACCESS_KEY_ID",
                "AWS_SECRET_ACCESS_KEY",
            }:
                github.update_secret(owner, repo, name, os.environ[name])
        elif config["provider"] == "gcp":
            github.update_secret(owner, repo, "PROJECT_ID", os.environ["PROJECT_ID"])
            with open(os.environ["GOOGLE_CREDENTIALS"]) as f:
                github.update_secret(owner, repo, "GOOGLE_CREDENTIALS", f.read())
        elif config["provider"] == "azure":
            for name in {
                "ARM_CLIENT_ID",
                "ARM_CLIENT_SECRET",
                "ARM_SUBSCRIPTION_ID",
                "ARM_TENANT_ID",
            }:
                github.update_secret(owner, repo, name, os.environ[name])
        github.update_secret(
            owner, repo, "REPOSITORY_ACCESS_TOKEN", os.environ["GITHUB_TOKEN"]
        )
    except requests.exceptions.HTTPError as he:
        raise ValueError(
            f"Unable to set Secrets on GitHub repo https://github.com/{owner}/{repo} - error message from GitHub is: {he}"
        )

    return f"git@github.com:{owner}/{repo}.git"


def git_repository_initialize(git_repository):
    if not git.is_git_repo("./"):
        git.initialize_git("./")
    git.add_git_remote(git_repository, path="./", remote_name="origin")


def auth0_auto_provision(config):
    auth0_config = create_client(config["domain"], config["project_name"])
    config["security"]["authentication"]["config"]["client_id"] = auth0_config[
        "client_id"
    ]
    config["security"]["authentication"]["config"]["client_secret"] = auth0_config[
        "client_secret"
    ]
    config["security"]["authentication"]["config"]["auth0_subdomain"] = auth0_config[
        "auth0_subdomain"
    ]



import pytest
import requests

from qhub.utils import AWS_ENV_DOCS, AZURE_ENV_DOCS, DO_ENV_DOCS, GCP_ENV_DOCS

LINKS_TO_TEST = [
    DO_ENV_DOCS,
    AWS_ENV_DOCS,
    GCP_ENV_DOCS,
    AZURE_ENV_DOCS,
]


@pytest.mark.parametrize("url,status_code", [(url, 200) for url in LINKS_TO_TEST])
def test_links(url, status_code):
    response = requests.get(url)
    assert response.status_code == status_code



import functools
import hashlib
import os
import pathlib
import shutil
import sys
from typing import Dict, List

from ruamel.yaml import YAML

from qhub.deprecate import DEPRECATED_FILE_PATHS
from qhub.provider.cicd.github import gen_qhub_linter, gen_qhub_ops
from qhub.provider.cicd.gitlab import gen_gitlab_ci
from qhub.stages import tf_objects


def render_template(output_directory, config_filename, force=False, dry_run=False):
    # get directory for qhub templates
    import qhub

    template_directory = pathlib.Path(qhub.__file__).parent / "template"

    # would be nice to remove assumption that input directory
    # is in local filesystem and a directory
    template_directory = pathlib.Path(template_directory)
    if not template_directory.is_dir():
        raise ValueError(f"template directory={template_directory} is not a directory")

    output_directory = pathlib.Path(output_directory).resolve()

    if output_directory == str(pathlib.Path.home()):
        print("ERROR: Deploying QHub in home directory is not advised!")
        sys.exit(1)

    # mkdir all the way down to repo dir so we can copy .gitignore
    # into it in remove_existing_renders
    output_directory.mkdir(exist_ok=True, parents=True)

    config_filename = pathlib.Path(config_filename)
    if not config_filename.is_file():
        raise ValueError(
            f"cookiecutter configuration={config_filename} is not filename"
        )

    with config_filename.open() as f:
        yaml = YAML(typ="safe", pure=True)
        config = yaml.load(f)

    # For any config values that start with
    # QHUB_SECRET_, set the values using the
    # corresponding env var.
    set_env_vars_in_config(config)

    config["repo_directory"] = output_directory.name
    config["qhub_config_yaml_path"] = str(config_filename.absolute())

    contents = render_contents(config)

    directories = [
        f"stages/02-infrastructure/{config['provider']}",
        "stages/03-kubernetes-initialize",
        "stages/04-kubernetes-ingress",
        "stages/05-kubernetes-keycloak",
        "stages/06-kubernetes-keycloak-configuration",
        "stages/07-kubernetes-services",
        "stages/08-qhub-tf-extensions",
    ]
    if (
        config["provider"] not in {"existing", "local"}
        and config["terraform_state"]["type"] == "remote"
    ):
        directories.append(f"stages/01-terraform-state/{config['provider']}")

    source_dirs = [os.path.join(str(template_directory), _) for _ in directories]
    output_dirs = [os.path.join(str(output_directory), _) for _ in directories]
    new, untracked, updated, deleted = inspect_files(
        source_dirs,
        output_dirs,
        source_base_dir=str(template_directory),
        output_base_dir=str(output_directory),
        ignore_filenames=[
            "terraform.tfstate",
            ".terraform.lock.hcl",
            "terraform.tfstate.backup",
        ],
        ignore_directories=[
            ".terraform",
            "__pycache__",
        ],
        deleted_paths=DEPRECATED_FILE_PATHS,
        contents=contents,
    )

    if new:
        print("The following files will be created:")
        for filename in sorted(new):
            print(f"   CREATED   {filename}")
    if updated:
        print("The following files will be updated:")
        for filename in sorted(updated):
            print(f"   UPDATED   {filename}")
    if deleted:
        print("The following files will be deleted:")
        for filename in sorted(deleted):
            print(f"   DELETED   {filename}")
    if untracked:
        print("The following files are untracked (only exist in output directory):")
        for filename in sorted(updated):
            print(f"   UNTRACKED {filename}")

    if dry_run:
        print("dry-run enabled no files will be created, updated, or deleted")
    else:
        for filename in new | updated:
            input_filename = os.path.join(str(template_directory), filename)
            output_filename = os.path.join(str(output_directory), filename)
            os.makedirs(os.path.dirname(output_filename), exist_ok=True)

            if os.path.exists(input_filename):
                shutil.copy(input_filename, output_filename)
            else:
                with open(output_filename, "w") as f:
                    f.write(contents[filename])

        for path in deleted:
            abs_path = os.path.abspath(os.path.join(str(output_directory), path))

            # be extra cautious that deleted path is within output_directory
            if not abs_path.startswith(str(output_directory)):
                raise Exception(
                    f"[ERROR] SHOULD NOT HAPPEN filename was about to be deleted but path={abs_path} is outside of output_directory"
                )

            if os.path.isfile(abs_path):
                os.remove(abs_path)
            elif os.path.isdir(abs_path):
                shutil.rmtree(abs_path)


def render_contents(config: Dict):
    """Dynamically generated contents from QHub configuration"""

    contents = {
        **tf_objects.stage_01_terraform_state(config),
        **tf_objects.stage_02_infrastructure(config),
        **tf_objects.stage_03_kubernetes_initialize(config),
        **tf_objects.stage_04_kubernetes_ingress(config),
        **tf_objects.stage_05_kubernetes_keycloak(config),
        **tf_objects.stage_06_kubernetes_keycloak_configuration(config),
        **tf_objects.stage_07_kubernetes_services(config),
        **tf_objects.stage_08_qhub_tf_extensions(config),
    }

    if config.get("ci_cd"):
        for fn, workflow in gen_cicd(config).items():
            contents.update(
                {
                    fn: workflow.json(
                        indent=2,
                        by_alias=True,
                        exclude_unset=True,
                        exclude_defaults=True,
                    )
                }
            )

    contents.update(gen_gitignore(config))

    return contents


def gen_gitignore(config):
    """
    Generate `.gitignore` file.
    Add files as needed.
    """

    from inspect import cleandoc

    filestoignore = """
        # ignore terraform state
        .terraform
        terraform.tfstate
        terraform.tfstate.backup
        .terraform.tfstate.lock.info

        # python
        __pycache__
    """
    return {".gitignore": cleandoc(filestoignore)}


def gen_cicd(config):
    """
    Use cicd schema to generate workflow files based on the
    `ci_cd` key in the `config`.

    For more detail on schema:
    GiHub-Actions - qhub/providers/cicd/github.py
    GitLab-CI - qhub/providers/cicd/gitlab.py
    """
    cicd_files = {}
    cicd_provider = config["ci_cd"]["type"]

    if cicd_provider == "github-actions":
        gha_dir = ".github/workflows/"
        cicd_files[gha_dir + "qhub-ops.yaml"] = gen_qhub_ops(config)
        cicd_files[gha_dir + "qhub-linter.yaml"] = gen_qhub_linter(config)

    elif cicd_provider == "gitlab-ci":
        cicd_files[".gitlab-ci.yml"] = gen_gitlab_ci(config)

    else:
        raise ValueError(
            f"The ci_cd provider, {cicd_provider}, is not supported. Supported providers include: `github-actions`, `gitlab-ci`."
        )

    return cicd_files


def inspect_files(
    source_dirs: str,
    output_dirs: str,
    source_base_dir: str,
    output_base_dir: str,
    ignore_filenames: List[str] = None,
    ignore_directories: List[str] = None,
    deleted_paths: List[str] = None,
    contents: Dict[str, str] = None,
):
    """Return created, updated and untracked files by computing a checksum over the provided directory

    Args:
        source_dirs (str): The source dir used as base for comparssion
        output_dirs (str): The destination dir which will be matched with
        source_base_dir (str): Relative base path to source directory
        output_base_dir (str): Relative base path to output directory
        ignore_filenames (list[str]): Filenames to ignore while comparing for changes
        ignore_directories (list[str]): Directories to ignore while comparing for changes
        deleted_paths (list[str]): Paths that if exist in output directory should be deleted
        contents (dict): filename to content mapping for dynamically generated files
    """
    ignore_filenames = ignore_filenames or []
    ignore_directories = ignore_directories or []
    contents = contents or {}

    source_files = {}
    output_files = {}

    def list_files(
        directory: str, ignore_filenames: List[str], ignore_directories: List[str]
    ):
        for root, dirs, files in os.walk(directory):
            dirs[:] = [d for d in dirs if d not in ignore_directories]
            for file in files:
                if file not in ignore_filenames:
                    yield os.path.join(root, file)

    for filename in contents:
        source_files[filename] = hashlib.sha256(
            contents[filename].encode("utf8")
        ).hexdigest()
        output_filename = os.path.join(output_base_dir, filename)
        if os.path.isfile(output_filename):
            output_files[filename] = hash_file(filename)

    deleted_paths = set()
    for path in deleted_paths:
        absolute_path = os.path.join(output_base_dir, path)
        if os.path.exists(absolute_path):
            deleted_paths.add(path)

    for source_dir, output_dir in zip(source_dirs, output_dirs):
        for filename in list_files(source_dir, ignore_filenames, ignore_directories):
            relative_path = os.path.relpath(filename, source_base_dir)
            source_files[relative_path] = hash_file(filename)

        for filename in list_files(output_dir, ignore_filenames, ignore_directories):
            relative_path = os.path.relpath(filename, output_base_dir)
            output_files[relative_path] = hash_file(filename)

    new_files = source_files.keys() - output_files.keys()
    untracted_files = output_files.keys() - source_files.keys()

    updated_files = set()
    for prevalent_file in source_files.keys() & output_files.keys():
        if source_files[prevalent_file] != output_files[prevalent_file]:
            updated_files.add(prevalent_file)

    return new_files, untracted_files, updated_files, deleted_paths


def hash_file(file_path: str):
    """Get the hex digest of the given file

    Args:
        file_path (str): path to file
    """
    with open(file_path, "rb") as f:
        return hashlib.sha256(f.read()).hexdigest()


def set_env_vars_in_config(config):
    """

    For values in the config starting with 'QHUB_SECRET_XXX' the environment
    variables are searched for the pattern XXX and the config value is
    modified. This enables setting secret values that should not be directly
    stored in the config file.

    NOTE: variables are most likely written to a file somewhere upon render. In
    order to further reduce risk of exposure of any of these variables you might
    consider preventing storage of the terraform render output.
    """
    private_entries = get_secret_config_entries(config)
    for idx in private_entries:
        set_qhub_secret(config, idx)


def get_secret_config_entries(config, config_idx=None, private_entries=None):
    output = private_entries or []
    if config_idx is None:
        sub_dict = config
        config_idx = []
    else:
        sub_dict = get_sub_config(config, config_idx)

    for key, value in sub_dict.items():
        if type(value) is dict:
            sub_dict_outputs = get_secret_config_entries(
                config, [*config_idx, key], private_entries
            )
            output = [*output, *sub_dict_outputs]
        else:
            if "QHUB_SECRET_" in str(value):
                output = [*output, [*config_idx, key]]
    return output


def get_sub_config(conf, conf_idx):
    sub_config = functools.reduce(dict.__getitem__, conf_idx, conf)
    return sub_config


def set_sub_config(conf, conf_idx, value):

    get_sub_config(conf, conf_idx[:-1])[conf_idx[-1]] = value


def set_qhub_secret(config, idx):
    placeholder = get_sub_config(config, idx)
    secret_var = get_qhub_secret(placeholder)
    set_sub_config(config, idx, secret_var)


def get_qhub_secret(secret_var):
    env_var = secret_var.lstrip("QHUB_SECRET_")
    val = os.environ.get(env_var)
    if not val:
        raise EnvironmentError(
            f"Since '{secret_var}' was found in the"
            " QHub config, the environment variable"
            f" '{env_var}' must be set."
        )
    return val



from qhub.constants import HIGHEST_SUPPORTED_K8S_VERSION


def filter_by_highest_supported_k8s_version(k8s_versions_list):
    filtered_k8s_versions_list = []
    for k8s_version in k8s_versions_list:
        if k8s_version <= HIGHEST_SUPPORTED_K8S_VERSION:
            filtered_k8s_versions_list.append(k8s_version)
    return filtered_k8s_versions_list



import json
import os
import tempfile
from urllib.parse import urlencode

from qhub.constants import DEFAULT_CONDA_STORE_IMAGE_TAG


def stage_01_terraform_state(stage_outputs, config):
    if config["provider"] == "do":
        return {
            "name": config["project_name"],
            "namespace": config["namespace"],
            "region": config["digital_ocean"]["region"],
        }
    elif config["provider"] == "gcp":
        return {
            "name": config["project_name"],
            "namespace": config["namespace"],
            "region": config["google_cloud_platform"]["region"],
        }
    elif config["provider"] == "aws":
        return {
            "name": config["project_name"],
            "namespace": config["namespace"],
        }
    elif config["provider"] == "azure":
        return {
            "name": config["project_name"],
            "namespace": config["namespace"],
            "region": config["azure"]["region"],
            "storage_account_postfix": config["azure"]["storage_account_postfix"],
            "state_resource_group_name": f'{config["project_name"]}-{config["namespace"]}-state',
        }
    else:
        return {}


def stage_02_infrastructure(stage_outputs, config):
    if config["provider"] == "local":
        return {
            "kubeconfig_filename": os.path.join(
                tempfile.gettempdir(), "QHUB_KUBECONFIG"
            ),
            "kube_context": config["local"].get("kube_context"),
        }
    elif config["provider"] == "existing":
        return {"kube_context": config["existing"].get("kube_context")}
    elif config["provider"] == "do":
        return {
            "name": config["project_name"],
            "environment": config["namespace"],
            "region": config["digital_ocean"]["region"],
            "kubernetes_version": config["digital_ocean"]["kubernetes_version"],
            "node_groups": config["digital_ocean"]["node_groups"],
            "kubeconfig_filename": os.path.join(
                tempfile.gettempdir(), "QHUB_KUBECONFIG"
            ),
            **config.get("do", {}).get("terraform_overrides", {}),
        }
    elif config["provider"] == "gcp":
        return {
            "name": config["project_name"],
            "environment": config["namespace"],
            "region": config["google_cloud_platform"]["region"],
            "project_id": config["google_cloud_platform"]["project"],
            "node_groups": [
                {
                    "name": key,
                    "instance_type": value["instance"],
                    "min_size": value["min_nodes"],
                    "max_size": value["max_nodes"],
                    "guest_accelerators": value["guest_accelerators"]
                    if "guest_accelerators" in value
                    else [],
                    **value,
                }
                for key, value in config["google_cloud_platform"]["node_groups"].items()
            ],
            "kubeconfig_filename": os.path.join(
                tempfile.gettempdir(), "QHUB_KUBECONFIG"
            ),
            **config.get("gcp", {}).get("terraform_overrides", {}),
        }
    elif config["provider"] == "azure":
        return {
            "name": config["project_name"],
            "environment": config["namespace"],
            "region": config["azure"]["region"],
            "kubernetes_version": config["azure"]["kubernetes_version"],
            "node_groups": config["azure"]["node_groups"],
            "kubeconfig_filename": os.path.join(
                tempfile.gettempdir(), "QHUB_KUBECONFIG"
            ),
            "resource_group_name": f'{config["project_name"]}-{config["namespace"]}',
            "node_resource_group_name": f'{config["project_name"]}-{config["namespace"]}-node-resource-group',
            **config.get("azure", {}).get("terraform_overrides", {}),
        }
    elif config["provider"] == "aws":
        return {
            "name": config["project_name"],
            "environment": config["namespace"],
            "node_groups": [
                {
                    "name": key,
                    "min_size": value["min_nodes"],
                    "desired_size": max(value["min_nodes"], 1),
                    "max_size": value["max_nodes"],
                    "gpu": value.get("gpu", False),
                    "instance_type": value["instance"],
                }
                for key, value in config["amazon_web_services"]["node_groups"].items()
            ],
            "kubeconfig_filename": os.path.join(
                tempfile.gettempdir(), "QHUB_KUBECONFIG"
            ),
            **config.get("aws", {}).get("terraform_overrides", {}),
        }
    else:
        return {}


def stage_03_kubernetes_initialize(stage_outputs, config):
    if config["provider"] == "gcp":
        gpu_enabled = any(
            node_group.get("guest_accelerators")
            for node_group in config["google_cloud_platform"]["node_groups"].values()
        )
        gpu_node_group_names = []

    elif config["provider"] == "aws":
        gpu_enabled = any(
            node_group.get("gpu")
            for node_group in config["amazon_web_services"]["node_groups"].values()
        )
        gpu_node_group_names = [
            group for group in config["amazon_web_services"]["node_groups"].keys()
        ]
    else:
        gpu_enabled = False
        gpu_node_group_names = []

    return {
        "name": config["project_name"],
        "environment": config["namespace"],
        "cloud-provider": config["provider"],
        "aws-region": config.get("amazon_web_services", {}).get("region"),
        "external_container_reg": config.get(
            "external_container_reg", {"enabled": False}
        ),
        "gpu_enabled": gpu_enabled,
        "gpu_node_group_names": gpu_node_group_names,
    }


def _calculate_note_groups(config):
    if config["provider"] == "aws":
        return {
            group: {"key": "eks.amazonaws.com/nodegroup", "value": group}
            for group in ["general", "user", "worker"]
        }
    elif config["provider"] == "gcp":
        return {
            group: {"key": "cloud.google.com/gke-nodepool", "value": group}
            for group in ["general", "user", "worker"]
        }
    elif config["provider"] == "azure":
        return {
            group: {"key": "azure-node-pool", "value": group}
            for group in ["general", "user", "worker"]
        }
    elif config["provider"] == "do":
        return {
            group: {"key": "doks.digitalocean.com/node-pool", "value": group}
            for group in ["general", "user", "worker"]
        }
    elif config["provider"] == "existing":
        return config["existing"].get("node_selectors")
    else:
        return config["local"]["node_selectors"]


def stage_04_kubernetes_ingress(stage_outputs, config):
    cert_type = config["certificate"]["type"]
    cert_details = {"certificate-service": cert_type}
    if cert_type == "lets-encrypt":
        cert_details["acme-email"] = config["certificate"]["acme_email"]
        cert_details["acme-server"] = config["certificate"]["acme_server"]

    return {
        **{
            "name": config["project_name"],
            "environment": config["namespace"],
            "node_groups": _calculate_note_groups(config),
            **config.get("ingress", {}).get("terraform_overrides", {}),
        },
        **cert_details,
    }


def stage_05_kubernetes_keycloak(stage_outputs, config):
    initial_root_password = (
        config["security"].get("keycloak", {}).get("initial_root_password", "")
    )
    if initial_root_password is None:
        initial_root_password = ""

    return {
        "name": config["project_name"],
        "environment": config["namespace"],
        "endpoint": config["domain"],
        "initial-root-password": initial_root_password,
        "overrides": [
            json.dumps(config["security"].get("keycloak", {}).get("overrides", {}))
        ],
        "node-group": _calculate_note_groups(config)["general"],
    }


def stage_06_kubernetes_keycloak_configuration(stage_outputs, config):
    realm_id = "qhub"

    users_group = (
        ["users"] if config["security"].get("shared_users_group", False) else []
    )

    return {
        "realm": realm_id,
        "realm_display_name": config["security"]
        .get("keycloak", {})
        .get("realm_display_name", realm_id),
        "authentication": config["security"]["authentication"],
        "keycloak_groups": ["admin", "developer", "analyst"] + users_group,
        "default_groups": ["analyst"] + users_group,
    }


def _split_docker_image_name(image_name):
    name, tag = image_name.split(":")
    return {"name": name, "tag": tag}


def stage_07_kubernetes_services(stage_outputs, config):
    final_logout_uri = f"https://{config['domain']}/hub/login"

    # Compound any logout URLs from extensions so they are are logged out in succession
    # when Keycloak and JupyterHub are logged out
    for ext in config.get("tf_extensions", []):
        if ext.get("logout", "") != "":
            final_logout_uri = "{}?{}".format(
                f"https://{config['domain']}/{ext['urlslug']}{ext['logout']}",
                urlencode({"redirect_uri": final_logout_uri}),
            )
    jupyterhub_theme = config["theme"]["jupyterhub"]
    if config["theme"]["jupyterhub"].get("display_version") and (
        not config["theme"]["jupyterhub"].get("version", False)
    ):
        jupyterhub_theme.update({"version": f"v{config['qhub_version']}"})

    return {
        "name": config["project_name"],
        "environment": config["namespace"],
        "endpoint": config["domain"],
        "realm_id": stage_outputs["stages/06-kubernetes-keycloak-configuration"][
            "realm_id"
        ]["value"],
        "node_groups": _calculate_note_groups(config),
        # conda-store
        "conda-store-environments": config["environments"],
        "conda-store-filesystem-storage": config["storage"]["conda_store"],
        "conda-store-extra-settings": config.get("conda_store", {}).get(
            "extra_settings", {}
        ),
        "conda-store-extra-config": config.get("conda_store", {}).get(
            "extra_config", ""
        ),
        "conda-store-image-tag": config.get("conda-store", {}).get(
            "image_tag", DEFAULT_CONDA_STORE_IMAGE_TAG
        ),
        # jupyterhub
        "cdsdashboards": config["cdsdashboards"],
        "jupyterhub-theme": jupyterhub_theme,
        "jupyterhub-image": _split_docker_image_name(
            config["default_images"]["jupyterhub"]
        ),
        "jupyterhub-shared-storage": config["storage"]["shared_filesystem"],
        "jupyterhub-shared-endpoint": stage_outputs["stages/02-infrastructure"]
        .get("nfs_endpoint", {})
        .get("value"),
        "jupyterlab-profiles": config["profiles"]["jupyterlab"],
        "jupyterlab-image": _split_docker_image_name(
            config["default_images"]["jupyterlab"]
        ),
        "jupyterhub-overrides": [
            json.dumps(config.get("jupyterhub", {}).get("overrides", {}))
        ],
        "jupyterhub-hub-extraEnv": json.dumps(
            config.get("jupyterhub", {})
            .get("overrides", {})
            .get("hub", {})
            .get("extraEnv", [])
        ),
        # dask-gateway
        "dask-worker-image": _split_docker_image_name(
            config["default_images"]["dask_worker"]
        ),
        "dask-gateway-profiles": config["profiles"]["dask_worker"],
        # monitoring
        "monitoring-enabled": config["monitoring"]["enabled"],
        # argo-worfklows
        "argo-workflows-enabled": config["argo_workflows"]["enabled"],
        "argo-workflows-overrides": [
            json.dumps(config.get("argo_workflows", {}).get("overrides", {}))
        ],
        # kbatch
        "kbatch-enabled": config["kbatch"]["enabled"],
        # prefect
        "prefect-enabled": config.get("prefect", {}).get("enabled", False),
        "prefect-token": config.get("prefect", {}).get("token", ""),
        "prefect-image": config.get("prefect", {}).get("image", ""),
        "prefect-overrides": config.get("prefect", {}).get("overrides", {}),
        # clearml
        "clearml-enabled": config.get("clearml", {}).get("enabled", False),
        "clearml-enable-forwardauth": config.get("clearml", {}).get(
            "enable_forward_auth", False
        ),
        "clearml-overrides": [
            json.dumps(config.get("clearml", {}).get("overrides", {}))
        ],
        "jupyterhub-logout-redirect-url": final_logout_uri,
    }


def stage_08_qhub_tf_extensions(stage_outputs, config):
    return {
        "environment": config["namespace"],
        "endpoint": config["domain"],
        "realm_id": stage_outputs["stages/06-kubernetes-keycloak-configuration"][
            "realm_id"
        ]["value"],
        "tf_extensions": config.get("tf_extensions", []),
        "qhub_config_yaml": config,
        "keycloak_qhub_bot_password": stage_outputs["stages/05-kubernetes-keycloak"][
            "keycloak_qhub_bot_password"
        ]["value"],
        "helm_extensions": config.get("helm_extensions", []),
    }



# remove after next kubespawner release past 1/20/2022
# https://github.com/jupyterhub/kubespawner/pull/558
import os

import kubernetes.client.models
from tornado import gen

kubernetes.client.models.V1EndpointPort = kubernetes.client.models.CoreV1EndpointPort

import z2jh  # noqa: E402
from kubespawner import KubeSpawner  # noqa: E402

cdsdashboards = z2jh.get_config("custom.cdsdashboards")
conda_store_environments = z2jh.get_config("custom.environments")


@gen.coroutine
def get_username_hook(spawner):
    auth_state = yield spawner.user.get_auth_state()
    username = auth_state["oauth_user"]["preferred_username"]

    spawner.environment.update(
        {
            "PREFERRED_USERNAME": username,
        }
    )


c.Spawner.pre_spawn_hook = get_username_hook

if cdsdashboards["enabled"]:
    from cdsdashboards.app import CDS_TEMPLATE_PATHS
    from cdsdashboards.builder.kubebuilder import KubeBuilder
    from cdsdashboards.hubextension import cds_extra_handlers
    from cdsdashboards.hubextension.spawners.variablekube import VariableKubeSpawner

    c.JupyterHub.allow_named_servers = True
    c.JupyterHub.extra_handlers.extend(cds_extra_handlers)
    c.JupyterHub.template_paths.extend(CDS_TEMPLATE_PATHS)
    c.JupyterHub.spawner_class = VariableKubeSpawner
    c.CDSDashboardsConfig.builder_class = KubeBuilder
    c.VariableMixin.default_presentation_cmd = [
        "python3",
        "-m",
        "jhsingle_native_proxy.main",
    ]
    c.JupyterHub.default_url = "/hub/home"

    # Force dashboard creator to select an instance size
    c.CDSDashboardsConfig.spawn_default_options = False

    c.CDSDashboardsConfig.conda_envs = [
        environment["name"] for _, environment in conda_store_environments.items()
    ]

    # TODO: make timeouts configurable
    c.VariableMixin.proxy_ready_timeout = 600
    c.VariableMixin.proxy_request_timeout = 600

    home_username = f"/home/{os.getenv('PREFERRED_USERNAME')}"
    c.CDSDashboardsConfig.extra_presentation_types = ["panel-serve"]
    c.VariableMixin.extra_presentation_launchers = {
        "panel-serve": {
            "args": [
                "python3",
                "{presentation_path}",
                "{--}port",
                "{port}",
                "{--}address",
                "{origin_host}",
            ],
            "debug_args": [],
            "env": {"PYTHONPATH": f"{home_username}/{{presentation_dirname}}"},
        }
    }

else:
    c.JupyterHub.allow_named_servers = False
    c.JupyterHub.spawner_class = KubeSpawner



import logging
import os

import CloudFlare

logger = logging.getLogger(__name__)


def update_record(zone_name, record_name, record_type, record_address):
    for variable in {"CLOUDFLARE_TOKEN"}:
        if variable not in os.environ:
            raise ValueError(
                f"Cloudflare required environment variable={variable} not defined"
            )

    cf = CloudFlare.CloudFlare(token=os.environ["CLOUDFLARE_TOKEN"])

    record = {
        "name": record_name,
        "type": record_type,
        "content": record_address,
        "ttl": 1,
        "proxied": False,
    }

    zone_id = None
    for zone in cf.zones.get():
        if zone["name"] == zone_name:
            zone_id = zone["id"]
            break
    else:
        raise ValueError(f"Cloudflare zone {zone_name} not found")

    existing_record = cf.zones.dns_records.get(
        zone_id, params={"name": f"{record_name}.{zone_name}", "type": record_type}
    )
    if existing_record:
        logger.info(
            f"record name={record_name} type={record_type} address={record_address} already exists updating"
        )
        cf.zones.dns_records.put(zone_id, existing_record[0]["id"], data=record)
    else:
        logger.info(
            f"record name={record_name} type={record_type} address={record_address} does not exists creating"
        )
        cf.zones.dns_records.post(zone_id, data=record)



import os


def pip_install_qhub(qhub_version: str) -> str:
    qhub_gh_branch = os.environ.get("QHUB_GH_BRANCH", None)
    pip_install = f"pip install qhub=={qhub_version}"
    # dev branches
    if qhub_gh_branch:
        pip_install = (
            f"pip install git+https://github.com/Quansight/qhub.git@{qhub_gh_branch}"
        )

    return pip_install



import argparse
import logging
import sys

from pydantic.error_wrappers import ValidationError

from qhub.cli.cost import create_cost_subcommand
from qhub.cli.deploy import create_deploy_subcommand
from qhub.cli.destroy import create_destroy_subcommand
from qhub.cli.initialize import create_init_subcommand
from qhub.cli.keycloak import create_keycloak_subcommand
from qhub.cli.render import create_render_subcommand
from qhub.cli.support import create_support_subcommand
from qhub.cli.upgrade import create_upgrade_subcommand
from qhub.cli.validate import create_validate_subcommand
from qhub.provider.terraform import TerraformException
from qhub.version import __version__


def cli(args):
    parser = argparse.ArgumentParser(description="QHub command line")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
        help="QHub version number",
    )
    parser.set_defaults(func=None)

    subparser = parser.add_subparsers(help=f"QHub - {__version__}")
    create_deploy_subcommand(subparser)
    create_render_subcommand(subparser)
    create_init_subcommand(subparser)
    create_validate_subcommand(subparser)
    create_destroy_subcommand(subparser)
    create_support_subcommand(subparser)
    create_upgrade_subcommand(subparser)
    create_keycloak_subcommand(subparser)
    create_cost_subcommand(subparser)

    args = parser.parse_args(args)

    if args.func is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    logging.basicConfig(level=logging.INFO)

    try:
        args.func(args)
    except ValidationError as valerr:
        sys.exit(
            "Error: The schema validation of the qhub-config.yaml failed."
            " The following error message may be helpful in determining what went wrong:\n\n"
            + str(valerr)
        )
    except ValueError as ve:
        sys.exit("\nProblem encountered: " + str(ve) + "\n")
    except TerraformException:
        sys.exit("\nProblem encountered: Terraform error\n")



import logging
import pathlib

from qhub.deploy import deploy_configuration
from qhub.render import render_template
from qhub.schema import verify
from qhub.utils import load_yaml

logger = logging.getLogger(__name__)


def create_deploy_subcommand(subparser):
    subparser = subparser.add_parser("deploy")
    subparser.add_argument(
        "-c", "--config", help="qhub configuration yaml file", required=True
    )
    subparser.add_argument("-o", "--output", default="./", help="output directory")
    subparser.add_argument(
        "--dns-provider",
        choices=["cloudflare"],
        help="dns provider to use for registering domain name mapping",
    )
    subparser.add_argument(
        "--skip-remote-state-provision",
        action="store_true",
        help="Skip terraform state deployment which is often required in CI once the terraform remote state bootstrapping phase is complete",
    )
    subparser.add_argument(
        "--dns-auto-provision",
        action="store_true",
        help="Attempt to automatically provision DNS. For Auth0 is requires environment variables AUTH0_DOMAIN, AUTH0_CLIENTID, AUTH0_CLIENT_SECRET",
    )
    subparser.add_argument(
        "--disable-prompt",
        action="store_true",
        help="Disable human intervention",
    )
    subparser.add_argument(
        "--disable-render",
        action="store_true",
        help="Disable auto-rendering in deploy stage",
    )
    subparser.set_defaults(func=handle_deploy)


def handle_deploy(args):
    config_filename = pathlib.Path(args.config)
    if not config_filename.is_file():
        raise ValueError(
            f"passed in configuration filename={config_filename} must exist"
        )

    config = load_yaml(config_filename)

    verify(config)

    if not args.disable_render:
        render_template(args.output, args.config, force=True)

    deploy_configuration(
        config,
        args.dns_provider,
        args.dns_auto_provision,
        args.disable_prompt,
        args.skip_remote_state_provision,
    )



import logging
import pathlib

from qhub.keycloak import do_keycloak

logger = logging.getLogger(__name__)


def create_keycloak_subcommand(subparser):
    subparser = subparser.add_parser("keycloak")
    subparser.add_argument("-c", "--config", help="qhub configuration", required=True)
    subparser.add_argument(
        "keycloak_action",
        nargs="+",
        help="`adduser <username> [password]` or `listusers`",
    )
    subparser.set_defaults(func=handle_keycloak)


def handle_keycloak(args):
    config_filename = pathlib.Path(args.config)
    if not config_filename.is_file():
        raise ValueError(
            f"passed in configuration filename={config_filename} must exist"
        )

    do_keycloak(config_filename, *args.keycloak_action)



import copy
import functools
import json
import os

import z2jh
from tornado import gen


def base_profile_home_mounts(username):
    """Configure the home directory mount for user

    Ensure that user directory exists and user has permissions to
    read/write/execute.

    """
    home_pvc_name = z2jh.get_config("custom.home-pvc")
    skel_mount = z2jh.get_config("custom.skel-mount")
    pvc_home_mount_path = "home/{username}"
    pod_home_mount_path = "/home/{username}"

    extra_pod_config = {
        "volumes": [
            {
                "name": "home",
                "persistentVolumeClaim": {
                    "claimName": home_pvc_name,
                },
            },
            {
                "name": "skel",
                "configMap": {
                    "name": skel_mount["name"],
                },
            },
        ]
    }

    extra_container_config = {
        "volumeMounts": [
            {
                "mountPath": pod_home_mount_path.format(username=username),
                "name": "home",
                "subPath": pvc_home_mount_path.format(username=username),
            },
        ]
    }

    MKDIR_OWN_DIRECTORY = (
        "mkdir -p /mnt/{path} && chmod 777 /mnt/{path} && cp -r /etc/skel/. /mnt/{path}"
    )
    command = MKDIR_OWN_DIRECTORY.format(
        path=pvc_home_mount_path.format(username=username)
    )
    init_containers = [
        {
            "name": "initialize-home-mount",
            "image": "busybox:1.31",
            "command": ["sh", "-c", command],
            "securityContext": {"runAsUser": 0},
            "volumeMounts": [
                {"mountPath": "/mnt", "name": "home"},
                {"mountPath": "/etc/skel", "name": "skel"},
            ],
        }
    ]
    return {
        "extra_pod_config": extra_pod_config,
        "extra_container_config": extra_container_config,
        "init_containers": init_containers,
    }


def base_profile_shared_mounts(groups):
    """Configure the group directory mounts for user

    Ensure that {shared}/{group} directory exists and user has
    permissions to read/write/execute. Kubernetes does not allow the
    same pvc to be a volume thus we must check that the home and share
    pvc are not the same for some operation.

    """
    home_pvc_name = z2jh.get_config("custom.home-pvc")
    shared_pvc_name = z2jh.get_config("custom.shared-pvc")

    pvc_shared_mount_path = "shared/{group}"
    pod_shared_mount_path = "/shared/{group}"

    extra_pod_config = {"volumes": []}
    if home_pvc_name != shared_pvc_name:
        extra_pod_config["volumes"].append(
            {"name": "shared", "persistentVolumeClaim": {"claimName": shared_pvc_name}}
        )

    extra_container_config = {
        "volumeMounts": [
            {
                "mountPath": pod_shared_mount_path.format(group=group),
                "name": "shared" if home_pvc_name != shared_pvc_name else "home",
                "subPath": pvc_shared_mount_path.format(group=group),
            }
            for group in groups
        ]
    }

    MKDIR_OWN_DIRECTORY = "mkdir -p /mnt/{path} && chmod 777 /mnt/{path}"
    command = " && ".join(
        [
            MKDIR_OWN_DIRECTORY.format(path=pvc_shared_mount_path.format(group=group))
            for group in groups
        ]
    )
    init_containers = [
        {
            "name": "initialize-shared-mounts",
            "image": "busybox:1.31",
            "command": ["sh", "-c", command],
            "securityContext": {"runAsUser": 0},
            "volumeMounts": [
                {
                    "mountPath": "/mnt",
                    "name": "shared" if home_pvc_name != shared_pvc_name else "home",
                }
            ],
        }
    ]
    return {
        "extra_pod_config": extra_pod_config,
        "extra_container_config": extra_container_config,
        "init_containers": init_containers,
    }


def profile_conda_store_mounts(username, groups):
    """Configure the conda_store environment directories mounts for
    user

    Ensure that {shared}/{group} directory exists and user has
    permissions to read/write/execute.

    """
    conda_store_pvc_name = z2jh.get_config("custom.conda-store-pvc")
    conda_store_mount = z2jh.get_config("custom.conda-store-mount")

    extra_pod_config = {
        "volumes": [
            {
                "name": "conda-store",
                "persistentVolumeClaim": {
                    "claimName": conda_store_pvc_name,
                },
            }
        ]
    }

    conda_store_namespaces = [username, "filesystem", "default"] + groups
    extra_container_config = {
        "volumeMounts": [
            {
                "mountPath": os.path.join(conda_store_mount, namespace),
                "name": "conda-store",
                "subPath": namespace,
            }
            for namespace in conda_store_namespaces
        ]
    }

    MKDIR_OWN_DIRECTORY = "mkdir -p /mnt/{path} && chmod 755 /mnt/{path}"
    command = " && ".join(
        [
            MKDIR_OWN_DIRECTORY.format(path=namespace)
            for namespace in conda_store_namespaces
        ]
    )
    init_containers = [
        {
            "name": "initialize-conda-store-mounts",
            "image": "busybox:1.31",
            "command": ["sh", "-c", command],
            "securityContext": {"runAsUser": 0},
            "volumeMounts": [
                {
                    "mountPath": "/mnt",
                    "name": "conda-store",
                }
            ],
        }
    ]
    return {
        "extra_pod_config": extra_pod_config,
        "extra_container_config": extra_container_config,
        "init_containers": init_containers,
    }


def base_profile_extra_mounts():
    extra_mounts = z2jh.get_config("custom.extra-mounts")

    extra_pod_config = {
        "volumes": [
            {
                "name": volume["name"],
                "persistentVolumeClaim": {"claimName": volume["name"]},
            }
            if volume["kind"] == "persistentvolumeclaim"
            else {"name": volume["name"], "configMap": {"name": volume["name"]}}
            for mount_path, volume in extra_mounts.items()
        ]
    }

    extra_container_config = {
        "volumeMounts": [
            {
                "name": volume["name"],
                "mountPath": mount_path,
            }
            for mount_path, volume in extra_mounts.items()
        ]
    }
    return {
        "extra_pod_config": extra_pod_config,
        "extra_container_config": extra_container_config,
    }


def configure_user(username, groups, uid=1000, gid=100):
    environment = {
        # nss_wrapper
        # https://cwrap.org/nss_wrapper.html
        "LD_PRELOAD": "libnss_wrapper.so",
        "NSS_WRAPPER_PASSWD": "/tmp/passwd",
        "NSS_WRAPPER_GROUP": "/tmp/group",
        # default files created will have 775 permissions
        "NB_UMASK": "0002",
        # set default shell to bash
        "SHELL": "/bin/bash",
        # set home directory to username
        "HOME": f"/home/{username}",
    }

    etc_passwd, etc_group = generate_nss_files(
        users=[{"username": username, "uid": uid, "gid": gid}],
        groups=[{"groupname": "users", "gid": gid}],
    )

    jupyter_config = json.dumps(
        {
            # nb_conda_kernels configuration
            # https://github.com/Anaconda-Platform/nb_conda_kernels
            "CondaKernelSpecManager": {"name_format": "{environment}"}
        }
    )

    # condarc to add all the namespaces user has access to
    condarc = json.dumps(
        {
            "envs_dirs": [
                f"/home/conda/{_}/envs"
                for _ in [
                    username,
                    "filesystem",
                    "default",
                ]
                + groups
            ]
        }
    )

    command = " && ".join(
        [
            # nss_wrapper
            # https://cwrap.org/nss_wrapper.html
            f"echo '{etc_passwd}' > /tmp/passwd",
            f"echo '{etc_group}' > /tmp/group",
            # mount the shared directories for user only if there are
            # shared folders (groups) that the user is a member of
            # else ensure that the `shared` folder symlink does not exist
            f"ln -sfn /shared /home/{username}/shared"
            if groups
            else f"rm -f /home/{username}/shared",
            # conda-store environment configuration
            f"printf '{condarc}' > /home/{username}/.condarc",
            # jupyter configuration
            f"mkdir -p /home/{username}/.jupyter && printf '{jupyter_config}' > /home/{username}/.jupyter/jupyter_config.json",
        ]
    )
    lifecycle_hooks = {"postStart": {"exec": {"command": ["/bin/sh", "-c", command]}}}

    extra_container_config = {
        "workingDir": f"/home/{username}",
    }

    return {
        "environment": environment,
        "lifecycle_hooks": lifecycle_hooks,
        "uid": uid,
        "gid": gid,
        "fs_gid": gid,
        "notebook_dir": f"/home/{username}",
        "extra_container_config": extra_container_config,
    }


def render_profile(profile, username, groups, keycloak_profilenames):
    """Render each profile for user

    If profile is not available for given username, groups returns
    None. Otherwise profile is transformed into kubespawner profile.

    {
        display_name: "<heading for profile>",
        slug: "<longer description of profile>"
        default: "<only one profile can be default>",
        kubespawner_override: {
            # https://jupyterhub-kubespawner.readthedocs.io/en/latest/spawner.html
            ...
        }
    }
    """
    access = profile.get("access", "all")

    if access == "yaml":
        # check that username or groups in allowed groups for profile
        # profile.groups and profile.users can be None or empty lists, or may not be members of profile at all
        user_not_in_users = username not in set(profile.get("users", []) or [])
        user_not_in_groups = (
            set(groups) & set(profile.get("groups", []) or [])
        ) == set()
        if user_not_in_users and user_not_in_groups:
            return None
    elif access == "keycloak":
        # Keycloak mapper should provide the 'jupyterlab_profiles' attribute from groups/user
        if profile.get("display_name", None) not in keycloak_profilenames:
            return None

    profile = copy.copy(profile)
    profile_kubespawner_override = profile.get("kubespawner_override")
    profile["kubespawner_override"] = functools.reduce(
        deep_merge,
        [
            base_profile_home_mounts(username),
            base_profile_shared_mounts(groups),
            profile_conda_store_mounts(username, groups),
            base_profile_extra_mounts(),
            configure_user(username, groups),
            profile_kubespawner_override,
        ],
        {},
    )

    # We need to merge any env vars from the spawner with any overrides from the profile
    # This is mainly to ensure JUPYTERHUB_ANYONE/GROUP is passed through from the spawner
    # to control dashboard access.
    envvars_fixed = {**(profile["kubespawner_override"].get("environment", {}))}

    def preserve_envvars(spawner):
        # This adds in JUPYTERHUB_ANYONE/GROUP rather than overwrite all env vars,
        # if set in the spawner for a dashboard to control access.
        return {**envvars_fixed, **spawner.environment}

    profile["kubespawner_override"]["environment"] = preserve_envvars

    return profile


@gen.coroutine
def render_profiles(spawner):
    # jupyterhub does not yet manage groups but it will soon
    # so for now we rely on auth_state from the keycloak
    # userinfo request to have the groups in the key
    # "auth_state.oauth_user.groups"
    auth_state = yield spawner.user.get_auth_state()
    spawner.log.error(str(auth_state))

    username = auth_state["oauth_user"]["preferred_username"]
    # only return the lowest level group name
    # e.g. /projects/myproj -> myproj
    # and /developers -> developers
    groups = [os.path.basename(_) for _ in auth_state["oauth_user"]["groups"]]
    spawner.log.error(f"user info: {username} {groups}")

    keycloak_profilenames = auth_state["oauth_user"].get("jupyterlab_profiles", [])

    # fetch available profiles and render additional attributes
    profile_list = z2jh.get_config("custom.profiles")
    return list(
        filter(
            None,
            [
                render_profile(p, username, groups, keycloak_profilenames)
                for p in profile_list
            ],
        )
    )


c.KubeSpawner.profile_list = render_profiles


# Utils
def deep_merge(d1, d2):
    """Deep merge two dictionaries.
    >>> value_1 = {
    'a': [1, 2],
    'b': {'c': 1, 'z': [5, 6]},
    'e': {'f': {'g': {}}},
    'm': 1,
    }

    >>> value_2 = {
        'a': [3, 4],
        'b': {'d': 2, 'z': [7]},
        'e': {'f': {'h': 1}},
        'm': [1],
    }

    >>> print(deep_merge(value_1, value_2))
    {'m': 1, 'e': {'f': {'g': {}, 'h': 1}}, 'b': {'d': 2, 'c': 1, 'z': [5, 6, 7]}, 'a': [1, 2, 3,  4]}
    """
    if isinstance(d1, dict) and isinstance(d2, dict):
        d3 = {}
        for key in d1.keys() | d2.keys():
            if key in d1 and key in d2:
                d3[key] = deep_merge(d1[key], d2[key])
            elif key in d1:
                d3[key] = d1[key]
            elif key in d2:
                d3[key] = d2[key]
        return d3
    elif isinstance(d1, list) and isinstance(d2, list):
        return [*d1, *d2]
    else:  # if they don't match use left one
        return d1


def generate_nss_files(users, groups):
    etc_passwd = []
    passwd_format = "{username}:x:{uid}:{gid}:{username}:/home/{username}:/bin/bash"
    for user in users:
        etc_passwd.append(passwd_format.format(**user))

    etc_group = []
    group_format = "{groupname}:x:{gid}:"
    for group in groups:
        etc_group.append(group_format.format(**group))

    return "\n".join(etc_passwd), "\n".join(etc_group)



import logging
import pathlib

from qhub.upgrade import do_upgrade

logger = logging.getLogger(__name__)


def create_upgrade_subcommand(subparser):
    subparser = subparser.add_parser("upgrade")
    subparser.add_argument("-c", "--config", help="qhub configuration", required=True)
    subparser.add_argument(
        "--attempt-fixes",
        action="store_true",
        help="Attempt to fix the config for any incompatibilities between your old and new QHub versions.",
    )
    subparser.set_defaults(func=handle_upgrade)


def handle_upgrade(args):
    config_filename = pathlib.Path(args.config)
    if not config_filename.is_file():
        raise ValueError(
            f"passed in configuration filename={config_filename} must exist"
        )

    do_upgrade(config_filename, args.attempt_fixes)



import os

QHUB_HOSTNAME = os.environ.get("QHUB_HOSTNAME", "github-actions.qhub.dev")
GATEWAY_ENDPOINT = "gateway"

KEYCLOAK_USERNAME = os.environ["KEYCLOAK_USERNAME"]
KEYCLOAK_PASSWORD = os.environ["KEYCLOAK_PASSWORD"]



"""
Start a cluster with Dask Gateway, print the dashboard link, and run some tasks.
"""
import dask
import dask_gateway
import distributed
from distributed import wait


def inc(x):
    return x + 1


def main():
    print(f"dask version         = {dask.__version__}")
    print(f"dask_gateway version = {dask_gateway.__version__}")
    print(f"distributed version  = {distributed.__version__}")

    gateway = dask_gateway.Gateway()
    options = gateway.cluster_options(use_local_defaults=False)

    print("Starting cluster")
    cluster = gateway.new_cluster(options)
    client = cluster.get_client()
    print("Dashboard at:", client.dashboard_link)

    cluster.scale(2)

    futures = client.map(inc, list(range(100)))
    _ = wait(futures)

    print("Closing cluster")
    cluster.close()


if __name__ == "__main__":
    main()



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field

from qhub.constants import LATEST_SUPPORTED_PYTHON_VERSION
from qhub.provider.cicd.common import pip_install_qhub


class GLCI_extras(BaseModel):
    # to allow for dynamic key names
    __root__: Union[str, float, int]


class GLCI_image(BaseModel):
    name: str
    entrypoint: Optional[str]


class GLCI_rules(BaseModel):
    if_: Optional[str] = Field(alias="if")
    changes: Optional[List[str]]

    class Config:
        allow_population_by_field_name = True


class GLCI_job(BaseModel):
    image: Optional[Union[str, GLCI_image]]
    variables: Optional[Dict[str, str]]
    before_script: Optional[List[str]]
    after_script: Optional[List[str]]
    script: List[str]
    rules: Optional[List[GLCI_rules]]


class GLCI(BaseModel):
    __root__: Dict[str, GLCI_job]


def gen_gitlab_ci(config):

    branch = config["ci_cd"]["branch"]
    commit_render = config["ci_cd"].get("commit_render", True)
    before_script = config["ci_cd"].get("before_script")
    after_script = config["ci_cd"].get("after_script")
    pip_install = pip_install_qhub(config["qhub_version"])

    render_vars = {
        "COMMIT_MSG": "qhub-config.yaml automated commit: {{ '$CI_COMMIT_SHA' }}",
    }

    script = [
        f"git checkout {branch}",
        f"{pip_install}",
        "qhub deploy --config qhub-config.yaml --disable-prompt --skip-remote-state-provision",
    ]

    commit_render_script = [
        "git config user.email 'qhub@quansight.com'",
        "git config user.name 'gitlab ci'",
        "git add .",
        "git diff --quiet && git diff --staged --quiet || (git commit -m '${COMMIT_MSG}'",
        f"git push origin {branch})",
    ]

    if commit_render:
        script += commit_render_script

    rules = [
        GLCI_rules(
            if_=f"$CI_COMMIT_BRANCH == '{branch}'",
            changes=["qhub-config.yaml"],
        )
    ]

    render_qhub = GLCI_job(
        image=f"python:{LATEST_SUPPORTED_PYTHON_VERSION}",
        variables=render_vars,
        before_script=before_script,
        after_script=after_script,
        script=script,
        rules=rules,
    )

    return GLCI(
        __root__={
            "render-qhub": render_qhub,
        }
    )



import logging
import os
import subprocess
import textwrap

from qhub.provider import terraform
from qhub.provider.dns.cloudflare import update_record
from qhub.stages import checks, input_vars, state_imports
from qhub.utils import (
    check_cloud_credentials,
    keycloak_provider_context,
    kubernetes_provider_context,
    timer,
)

logger = logging.getLogger(__name__)


def provision_01_terraform_state(stage_outputs, config):
    directory = "stages/01-terraform-state"

    if config["provider"] in {"existing", "local"}:
        stage_outputs[directory] = {}
    else:
        stage_outputs[directory] = terraform.deploy(
            terraform_import=True,
            directory=os.path.join(directory, config["provider"]),
            input_vars=input_vars.stage_01_terraform_state(stage_outputs, config),
            state_imports=state_imports.stage_01_terraform_state(stage_outputs, config),
        )


def provision_02_infrastructure(stage_outputs, config, check=True):
    """Generalized method to provision infrastructure

    After successful deployment the following properties are set on
    `stage_outputs[directory]`.
      - `kubernetes_credentials` which are sufficient credentials to
        connect with the kubernetes provider
      - `kubeconfig_filename` which is a path to a kubeconfig that can
        be used to connect to a kubernetes cluster
      - at least one node running such that resources in the
        node_group.general can be scheduled

    At a high level this stage is expected to provision a kubernetes
    cluster on a given provider.
    """
    directory = "stages/02-infrastructure"

    stage_outputs[directory] = terraform.deploy(
        os.path.join(directory, config["provider"]),
        input_vars=input_vars.stage_02_infrastructure(stage_outputs, config),
    )

    if check:
        checks.stage_02_infrastructure(stage_outputs, config)


def provision_03_kubernetes_initialize(stage_outputs, config, check=True):
    directory = "stages/03-kubernetes-initialize"

    stage_outputs[directory] = terraform.deploy(
        directory=directory,
        input_vars=input_vars.stage_03_kubernetes_initialize(stage_outputs, config),
    )

    if check:
        checks.stage_03_kubernetes_initialize(stage_outputs, config)


def provision_04_kubernetes_ingress(stage_outputs, config, check=True):
    directory = "stages/04-kubernetes-ingress"

    stage_outputs[directory] = terraform.deploy(
        directory=directory,
        input_vars=input_vars.stage_04_kubernetes_ingress(stage_outputs, config),
    )

    if check:
        checks.stage_04_kubernetes_ingress(stage_outputs, config)


def add_clearml_dns(zone_name, record_name, record_type, ip_or_hostname):
    dns_records = [
        f"app.clearml.{record_name}",
        f"api.clearml.{record_name}",
        f"files.clearml.{record_name}",
    ]

    for dns_record in dns_records:
        update_record(zone_name, dns_record, record_type, ip_or_hostname)


def provision_ingress_dns(
    stage_outputs,
    config,
    dns_provider: str,
    dns_auto_provision: bool,
    disable_prompt: bool = True,
    check: bool = True,
):
    directory = "stages/04-kubernetes-ingress"

    ip_or_name = stage_outputs[directory]["load_balancer_address"]["value"]
    ip_or_hostname = ip_or_name["hostname"] or ip_or_name["ip"]

    if dns_auto_provision and dns_provider == "cloudflare":
        record_name, zone_name = (
            config["domain"].split(".")[:-2],
            config["domain"].split(".")[-2:],
        )
        record_name = ".".join(record_name)
        zone_name = ".".join(zone_name)
        if config["provider"] in {"do", "gcp", "azure"}:
            update_record(zone_name, record_name, "A", ip_or_hostname)
            if config.get("clearml", {}).get("enabled"):
                add_clearml_dns(zone_name, record_name, "A", ip_or_hostname)

        elif config["provider"] == "aws":
            update_record(zone_name, record_name, "CNAME", ip_or_hostname)
            if config.get("clearml", {}).get("enabled"):
                add_clearml_dns(zone_name, record_name, "CNAME", ip_or_hostname)
        else:
            logger.info(
                f"Couldn't update the DNS record for cloud provider: {config['provider']}"
            )
    elif not disable_prompt:
        input(
            f"Take IP Address {ip_or_hostname} and update DNS to point to "
            f'"{config["domain"]}" [Press Enter when Complete]'
        )

    if check:
        checks.check_ingress_dns(stage_outputs, config, disable_prompt)


def provision_05_kubernetes_keycloak(stage_outputs, config, check=True):
    directory = "stages/05-kubernetes-keycloak"

    stage_outputs[directory] = terraform.deploy(
        directory=directory,
        input_vars=input_vars.stage_05_kubernetes_keycloak(stage_outputs, config),
    )

    if check:
        checks.stage_05_kubernetes_keycloak(stage_outputs, config)


def provision_06_kubernetes_keycloak_configuration(stage_outputs, config, check=True):
    directory = "stages/06-kubernetes-keycloak-configuration"

    stage_outputs[directory] = terraform.deploy(
        directory=directory,
        input_vars=input_vars.stage_06_kubernetes_keycloak_configuration(
            stage_outputs, config
        ),
    )

    if check:
        checks.stage_06_kubernetes_keycloak_configuration(stage_outputs, config)


def provision_07_kubernetes_services(stage_outputs, config, check=True):
    directory = "stages/07-kubernetes-services"

    stage_outputs[directory] = terraform.deploy(
        directory=directory,
        input_vars=input_vars.stage_07_kubernetes_services(stage_outputs, config),
    )

    if check:
        checks.stage_07_kubernetes_services(stage_outputs, config)


def provision_08_qhub_tf_extensions(stage_outputs, config, check=True):
    directory = "stages/08-qhub-tf-extensions"

    stage_outputs[directory] = terraform.deploy(
        directory=directory,
        input_vars=input_vars.stage_08_qhub_tf_extensions(stage_outputs, config),
    )

    if check:
        pass


def guided_install(
    config,
    dns_provider,
    dns_auto_provision,
    disable_prompt=False,
    skip_remote_state_provision=False,
):
    # 01 Check Environment Variables
    check_cloud_credentials(config)

    stage_outputs = {}
    if (
        config["provider"] not in {"existing", "local"}
        and config["terraform_state"]["type"] == "remote"
    ):
        if skip_remote_state_provision:
            print("Skipping remote state provision")
        else:
            provision_01_terraform_state(stage_outputs, config)

    provision_02_infrastructure(stage_outputs, config)

    with kubernetes_provider_context(
        stage_outputs["stages/02-infrastructure"]["kubernetes_credentials"]["value"]
    ):
        provision_03_kubernetes_initialize(stage_outputs, config)
        provision_04_kubernetes_ingress(stage_outputs, config)
        provision_ingress_dns(
            stage_outputs,
            config,
            dns_provider=dns_provider,
            dns_auto_provision=dns_auto_provision,
            disable_prompt=disable_prompt,
        )
        provision_05_kubernetes_keycloak(stage_outputs, config)

        with keycloak_provider_context(
            stage_outputs["stages/05-kubernetes-keycloak"]["keycloak_credentials"][
                "value"
            ]
        ):
            provision_06_kubernetes_keycloak_configuration(stage_outputs, config)
            provision_07_kubernetes_services(stage_outputs, config)
            provision_08_qhub_tf_extensions(stage_outputs, config)

            print("QHub deployed successfully")

    print("Services:")
    for service_name, service in stage_outputs["stages/07-kubernetes-services"][
        "service_urls"
    ]["value"].items():
        print(f" - {service_name} -> {service['url']}")

    print(
        f"Kubernetes kubeconfig located at file://{stage_outputs['stages/02-infrastructure']['kubeconfig_filename']['value']}"
    )
    username = "root"
    password = (
        config.get("security", {}).get("keycloak", {}).get("initial_root_password", "")
    )
    if password:
        print(f"Kubecloak master realm username={username} password={password}")

    print(
        "Additional administration docs can be found at https://docs.qhub.dev/en/stable/source/admin_guide/"
    )


def deploy_configuration(
    config,
    dns_provider,
    dns_auto_provision,
    disable_prompt,
    skip_remote_state_provision,
):
    if config.get("prevent_deploy", False):
        # Note if we used the Pydantic model properly, we might get that qhub_config.prevent_deploy always exists but defaults to False
        raise ValueError(
            textwrap.dedent(
                """
        Deployment prevented due to the prevent_deploy setting in your qhub-config.yaml file.
        You could remove that field to deploy your QHub, but please do NOT do so without fully understanding why that value was set in the first place.

        It may have been set during an upgrade of your qhub-config.yaml file because we do not believe it is safe to redeploy the new
        version of QHub without having a full backup of your system ready to restore. It may be known that an in-situ upgrade is impossible
        and that redeployment will tear down your existing infrastructure before creating an entirely new QHub without your old data.

        PLEASE get in touch with Quansight at https://github.com/Quansight/qhub for assistance in proceeding.
        Your data may be at risk without our guidance.
        """
            )
        )

    logger.info(f'All qhub endpoints will be under https://{config["domain"]}')

    with timer(logger, "deploying QHub"):
        try:
            guided_install(
                config,
                dns_provider,
                dns_auto_provision,
                disable_prompt,
                skip_remote_state_provision,
            )
        except subprocess.CalledProcessError as e:
            logger.error(e.output)
            raise e



import argparse
import logging
import pathlib
import time

from qhub.utils import check_cloud_credentials, load_yaml, timer

logging.basicConfig(level=logging.INFO)


def main():
    parser = argparse.ArgumentParser(description="Force Destroy AWS environment.")
    parser.add_argument("-c", "--config", help="qhub configuration", required=True)
    args = parser.parse_args()

    handle_force_destroy(args)


def handle_force_destroy(args):
    config_filename = pathlib.Path(args.config)
    if not config_filename.is_file():
        raise ValueError(
            f"passed in configuration filename={config_filename} must exist"
        )

    config = load_yaml(config_filename)

    # Don't verify(config) in case the schema has changed - just pick out the important bits and tear down

    force_destroy_configuration(config)


def parse_arn(arn):
    # http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html
    elements = arn.split(":", 5)
    result = {
        # "arn": elements[0],
        "partition": elements[1],
        "service": elements[2],
        "region": elements[3],
        "account": elements[4],
        "resource": elements[5],
        "resource_type": None,
        "arn": arn,  # Full ARN
    }
    if "/" in result["resource"]:
        result["resource_type"], result["resource"] = result["resource"].split("/", 1)
    elif ":" in result["resource"]:
        result["resource_type"], result["resource"] = result["resource"].split(":", 1)
    return result


def force_destroy_configuration(config):
    logging.info("""FORCE Removing all infrastructure (not using terraform).""")

    with timer(logging, "destroying QHub"):
        # 01 Check we have cloud details we need
        check_cloud_credentials(config)

        if config.get("provider", "") != "aws":
            raise ValueError("force-destroy currently only available for AWS")

        project_name = config.get("project_name", "").strip()

        if project_name == "":
            raise ValueError("project_name cannot be blank")

        if "amazon_web_services" not in config:
            raise ValueError(
                "amazon_web_services section must exist in qhub-config.yaml"
            )

        region = config["amazon_web_services"].get("region", "").strip()

        if region == "":
            raise ValueError(
                "amazon_web_services.region must exist in qhub-config.yaml"
            )

        logging.info(f"Remove AWS project {project_name} in region {region}")

        env = config.get("namespace", "dev").strip()

        # 02 Remove all infrastructure
        try:
            import boto3
        except ImportError:
            raise ValueError(
                "Please ensure boto3 package is installed using: pip install boto3==1.17.98"
            )

        restag = boto3.client("resourcegroupstaggingapi", region_name=region)

        filter_params = dict(
            TagFilters=[
                {
                    "Key": "Owner",
                    "Values": [
                        "terraform",
                        "terraform-state",
                    ],
                },
                {
                    "Key": "Environment",
                    "Values": [
                        env,
                    ],
                },
                {
                    "Key": "Project",
                    "Values": [
                        project_name,
                    ],
                },
            ],
            ResourcesPerPage=50,
        )

        resources = []

        response = restag.get_resources(**filter_params)

        resources.extend(response["ResourceTagMappingList"])

        while "PaginationToken" in response and response["PaginationToken"]:
            token = response["PaginationToken"]
            response = restag.get_resources(**filter_params, PaginationToken=token)
            resources.extend(response["ResourceTagMappingList"])

        # Load Balancer and other K8s-generated resources will need to be queried separately:

        filter_params = dict(
            TagFilters=[
                {
                    "Key": f"kubernetes.io/cluster/{project_name}-{env}",
                    "Values": [
                        "owned",
                    ],
                }
            ],
            ResourcesPerPage=50,
        )

        response = restag.get_resources(**filter_params)
        resources.extend(response["ResourceTagMappingList"])

        # IAM

        iam = boto3.resource("iam")
        for suffix in ("eks-cluster-role", "eks-node-group-role"):

            try:
                role = iam.Role(f"{project_name}-{env}-{suffix}")

                if role.tags is not None:

                    tags_dict = dict(
                        [(t["Key"], t.get("Value", "")) for t in role.tags]
                    )

                    if (
                        tags_dict.get("Owner", "") == "terraform"
                        and tags_dict.get("Environment", "") == env
                        and tags_dict.get("Project", "") == project_name
                    ):
                        resources.append({"ResourceARN": role.arn})

            except iam.meta.client.exceptions.NoSuchEntityException:
                pass

        # Summarize resources

        type_groups = {}
        for r in resources:
            de_arned = parse_arn(r["ResourceARN"])
            t = f"{de_arned['service']}-{de_arned['resource_type']}"
            type_groups.setdefault(t, []).append(de_arned)
            logging.info(r["ResourceARN"])

        logging.info([(k, len(v)) for k, v in type_groups.items()])

        # Order
        priority_types = (
            "eks-nodegroup",
            "eks-cluster",
            "elasticloadbalancing-loadbalancer",
            "ec2-internet-gateway",
            "ec2-route-table",
            "elasticfilesystem-file-system",
            "ec2-subnet",
            "ec2-security-group",
            "ec2-vpc",
            "ecr-repository",
            "dynamodb-table",
            "s3-None",
            "resource-groups-group",
            "iam-role",
        )

        for pt in priority_types:
            logging.info(f"Inspect {pt}")
            for r in type_groups.get(pt, []):
                if pt == "eks-nodegroup":
                    nodegroup_resource = r["resource"].split("/")

                    cluster_name = nodegroup_resource[0]
                    nodegroup_name = nodegroup_resource[1]

                    logging.info(f"Delete {nodegroup_name} on cluster {cluster_name}")

                    client = boto3.client("eks", region_name=region)
                    client.delete_nodegroup(
                        clusterName=cluster_name, nodegroupName=nodegroup_name
                    )

                elif pt == "eks-cluster":
                    logging.info(f"Delete EKS cluster {r['resource']}")

                    client = boto3.client("eks", region_name=region)

                    response = client.list_nodegroups(clusterName=r["resource"])
                    while len(response["nodegroups"]) > 0:
                        logging.info("Nodegroups still present, sleep 10")
                        time.sleep(10)
                        response = client.list_nodegroups(clusterName=r["resource"])

                    client.delete_cluster(name=r["resource"])

                elif pt == "elasticloadbalancing-loadbalancer":
                    client = boto3.client("elb", region_name=region)

                    logging.info(f"Inspect Load balancer {r['resource']}")

                    logging.info(f"Delete Load balancer {r['resource']}")
                    response = client.delete_load_balancer(
                        LoadBalancerName=r["resource"]
                    )

                elif pt == "ec2-route-table":
                    logging.info(f"Inspect route table {r['resource']}")
                    ec2 = boto3.resource("ec2", region_name=region)
                    route_table = ec2.RouteTable(r["resource"])

                    for assoc in route_table.associations:
                        logging.info(f"Delete route table assoc {assoc.id}")
                        assoc.delete()

                    time.sleep(10)

                    logging.info(f"Delete route table {r['resource']}")
                    route_table.delete()

                elif pt == "ec2-subnet":
                    logging.info(f"Inspect subnet {r['resource']}")
                    ec2 = boto3.resource("ec2", region_name=region)
                    subnet = ec2.Subnet(r["resource"])

                    for ni in subnet.network_interfaces.all():
                        ni.load()
                        # But can only detach if attached...
                        if ni.attachment:
                            ni.detach(DryRun=False, Force=True)
                            ni.delete()

                    logging.info(f"Delete subnet {r['resource']}")
                    subnet.delete(DryRun=False)

                elif pt == "ec2-security-group":
                    logging.info(f"Inspect security group {r['resource']}")
                    ec2 = boto3.resource("ec2", region_name=region)
                    security_group = ec2.SecurityGroup(r["resource"])

                    for ipperms in security_group.ip_permissions_egress:
                        security_group.revoke_egress(
                            DryRun=False, IpPermissions=[ipperms]
                        )

                    for ipperms in security_group.ip_permissions:
                        security_group.revoke_ingress(
                            DryRun=False, IpPermissions=[ipperms]
                        )

                    logging.info(f"Delete security group {r['resource']}")
                    security_group.delete(DryRun=False)

                elif pt == "ec2-internet-gateway":
                    logging.info(f"Inspect internet gateway {r['resource']}")

                    ec2 = boto3.resource("ec2", region_name=region)
                    internet_gateway = ec2.InternetGateway(r["resource"])

                    for attach in internet_gateway.attachments:
                        logging.info(f"Inspect IG attachment {attach['VpcId']}")
                        if attach.get("State", "") == "available":
                            logging.info(f"Detach from VPC {attach['VpcId']}")
                            internet_gateway.detach_from_vpc(VpcId=attach["VpcId"])

                    time.sleep(10)

                    logging.info(f"Delete internet gateway {r['resource']}")
                    internet_gateway.delete(DryRun=False)

                elif pt == "elasticfilesystem-file-system":
                    client = boto3.client("efs", region_name=region)

                    logging.info(f"Delete efs {r['resource']}")

                    mts = client.describe_mount_targets(FileSystemId=r["resource"])

                    for mt in mts["MountTargets"]:
                        client.delete_mount_target(MountTargetId=mt["MountTargetId"])

                    response = client.delete_file_system(FileSystemId=r["resource"])

                    ## Should wait until this returns botocore.errorfactory.FileSystemNotFound:
                    # response = client.describe_file_systems(
                    #    FileSystemId=r['resource']
                    # )

                elif pt == "ec2-vpc":
                    logging.info(f"Inspect VPC {r['resource']}")

                    ec2 = boto3.resource("ec2", region_name=region)

                    vpc = ec2.Vpc(r["resource"])

                    # for cidr_assoc in vpc.cidr_block_association_set:
                    #    logging.info(cidr_assoc)
                    #    r = vpc.disassociate_subnet_cidr_block(
                    #        AssociationId=cidr_assoc['AssociationId']
                    #    )
                    #    logging.info(r)

                    logging.info(f"Delete VPC {r['resource']}")
                    vpc.delete()

                elif pt == "ecr-repository":
                    logging.info(f"Inspect ECR {r['resource']}")
                    client = boto3.client("ecr", region_name=region)

                    logging.info(f"Delete ecr {r['account']} / {r['resource']}")

                    response = response = client.delete_repository(
                        registryId=r["account"],
                        repositoryName=r["resource"],
                        force=True,
                    )

                elif pt == "s3-None":
                    logging.info(f"Inspect S3 {r['resource']}")
                    s3 = boto3.resource("s3", region_name=region)

                    logging.info(f"Delete s3 {r['resource']}")

                    bucket = s3.Bucket(r["resource"])

                    r = bucket.objects.all().delete()

                    r = bucket.object_versions.delete()

                    response = bucket.delete()

                elif pt == "dynamodb-table":
                    logging.info(f"Inspect DynamoDB {r['resource']}")

                    client = boto3.client("dynamodb", region_name=region)

                    logging.info(f"Delete DynamoDB {r['resource']}")

                    response = client.delete_table(TableName=r["resource"])

                elif pt == "resource-groups-group":
                    logging.info(f"Inspect Resource Group {r['resource']}")

                    client = boto3.client("resource-groups", region_name=region)

                    logging.info(f"Delete Resource Group {r['resource']}")

                    response = client.delete_group(Group=r["arn"])

                elif pt == "iam-role":
                    logging.info(f"Inspect IAM Role {r['resource']}")
                    iam = boto3.resource("iam")
                    role = iam.Role(r["resource"])

                    for policy in role.attached_policies.all():
                        logging.info(f"Detach Role policy {policy.arn}")
                        response = role.detach_policy(PolicyArn=policy.arn)

                    logging.info(f"Delete IAM Role {r['resource']}")
                    role.delete()


if __name__ == "__main__":
    main()



TERRAFORM_VERSION = "1.0.5"

HIGHEST_SUPPORTED_K8S_VERSION = "1.23.5"

DEFAULT_QHUB_DASK_VERSION = "0.4.3"
DEFAULT_QHUB_IMAGE_TAG = "main"

DEFAULT_CONDA_STORE_IMAGE_TAG = "v0.4.11"

LATEST_SUPPORTED_PYTHON_VERSION = "3.10"



import json


def dask_gateway_config(path="/var/lib/dask-gateway/config.json"):
    with open(path) as f:
        return json.load(f)


config = dask_gateway_config()

c.KubeController.address = ":8000"
c.KubeController.api_url = f'http://{config["gateway_service_name"]}.{config["gateway_service_namespace"]}:8000/api'
c.KubeController.gateway_instance = config["gateway_service_name"]
c.KubeController.proxy_prefix = config["gateway"]["prefix"]
c.KubeController.proxy_web_middlewares = [
    {
        "name": config["gateway_cluster_middleware_name"],
        "namespace": config["gateway_cluster_middleware_namespace"],
    }
]
c.KubeController.log_level = config["controller"]["loglevel"]
c.KubeController.completed_cluster_max_age = config["controller"][
    "completedClusterMaxAge"
]
c.KubeController.completed_cluster_cleanup_period = config["controller"][
    "completedClusterCleanupPeriod"
]
c.KubeController.backoff_base_delay = config["controller"]["backoffBaseDelay"]
c.KubeController.backoff_max_delay = config["controller"]["backoffMaxDelay"]
c.KubeController.k8s_api_rate_limit = config["controller"]["k8sApiRateLimit"]
c.KubeController.k8s_api_rate_limit_burst = config["controller"]["k8sApiRateLimitBurst"]

c.KubeController.proxy_web_entrypoint = "websecure"
c.KubeController.proxy_tcp_entrypoint = "tcp"



from nox import session


@session(reuse_venv=True)
def docs(session):
    session.install("-e.")
    session.install("-rdocs/requirements.txt")
    session.run("sphinx-build", "docs", "_build/html")
